# Copyright (c) 2025 Cloudification GmbH.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging

from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as oscutils

from manilaclient.common._i18n import _
from manilaclient.common.apiclient import utils as apiutils
from manilaclient.common import constants
from manilaclient.osc import utils

LOG = logging.getLogger(__name__)

ATTRIBUTES = [
    'id',
    'name',
    'description',
    'specs',
    'created_at',
    'updated_at',
]


def format_qos_type(qos_type, formatter='table'):
    specs = qos_type.specs
    if formatter == 'table':
        qos_type._info.update({'specs': utils.format_properties(specs)})
    else:
        qos_type._info.update({'specs': specs})
    return qos_type


class CreateQosType(command.ShowOne):
    """Create new qos type."""

    _description = _("Create new qos type")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar="<name>",
            help=_('QoS type name. This must be unique.'),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            default=None,
            help=_("QoS type description."),
        )
        parser.add_argument(
            "--spec",
            type=str,
            metavar='<key=value>',
            action='append',
            default=None,
            help=_(
                "Spec key and value of QoS type that will be"
                " used for QoS type creation. OPTIONAL: Default=None."
                " Example: --spec qos_type='fixed' --spec peak_iops=300."
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        kwargs = {'name': parsed_args.name}
        if parsed_args.description:
            kwargs['description'] = parsed_args.description

        if parsed_args.spec:
            specs = utils.extract_properties(parsed_args.spec)
            kwargs['specs'] = specs

        qos_type = share_client.qos_types.create(**kwargs)
        formatted_type = format_qos_type(qos_type, parsed_args.formatter)

        return (
            ATTRIBUTES,
            oscutils.get_dict_properties(formatted_type._info, ATTRIBUTES),
        )


class DeleteQosType(command.Command):
    """Delete a qos type."""

    _description = _("Delete a qos type")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_types',
            metavar="<qos_types>",
            nargs="+",
            help=_("Name or ID of the qos type(s) to delete"),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share
        result = 0

        for qos_type in parsed_args.qos_types:
            try:
                qos_type_obj = apiutils.find_resource(
                    share_client.qos_types, qos_type
                )

                share_client.qos_types.delete(qos_type_obj)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete qos type with "
                        "name or ID '%(qos_type)s': %(e)s"
                    ),
                    {'qos_type': qos_type, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.qos_types)
            msg = _("%(result)s of %(total)s qos types failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class SetQosType(command.Command):
    """Set qos type description or specs."""

    _description = _("Set qos type description or specs")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_type',
            metavar="<qos_type>",
            help=_("Name or ID of the qos type to modify"),
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            default=None,
            help=_("New description of qos type."),
        )
        parser.add_argument(
            "--spec",
            type=str,
            metavar='<key=value>',
            action='append',
            default=None,
            help=_(
                "Spec key and value of qos type that will be "
                "used for QoS type. OPTIONAL: Default=None. For "
                "example --spec qos_type='fixed' --spec peak_iops=300"
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        qos_type = apiutils.find_resource(
            share_client.qos_types, parsed_args.qos_type
        )

        kwargs = {}
        if parsed_args.description:
            kwargs['description'] = parsed_args.description
        if kwargs:
            try:
                qos_type.update(**kwargs)
            except Exception as e:
                raise exceptions.CommandError(
                    _("Failed to set qos type description: %s") % e
                )

        # These are dict of key=value to be added as qos type specs.
        if parsed_args.spec:
            specs = utils.extract_properties(parsed_args.spec)
            try:
                qos_type.set_keys(specs)
            except Exception as e:
                raise exceptions.CommandError(
                    _("Failed to set qos type spec: %s") % e
                )


class UnsetQosType(command.Command):
    """Unset qos type specs."""

    _description = _("Unset qos type specs")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_type',
            metavar="<qos_type>",
            help=_("Name or ID of the qos type to modify"),
        )
        parser.add_argument(
            "--description",
            action='store_true',
            help=_("Unset qos type description."),
        )
        parser.add_argument(
            '--spec',
            metavar='<key>',
            action='append',
            help=_('Remove specified spec from this qos type'),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        qos_type = apiutils.find_resource(
            share_client.qos_types, parsed_args.qos_type
        )

        kwargs = {}
        if parsed_args.description:
            kwargs['description'] = None
        if kwargs:
            try:
                qos_type.update(**kwargs)
            except Exception as e:
                raise exceptions.CommandError(
                    _("Failed to unset qos type description: %s") % e
                )

        # These are list of keys to be deleted from qos type specs.
        if parsed_args.spec:
            try:
                qos_type.unset_keys(parsed_args.spec)
            except Exception as e:
                raise exceptions.CommandError(
                    _("Failed to remove qos type spec: %s") % e
                )


class ListQosType(command.Lister):
    """List Qos Types."""

    _description = _("List qos types")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--name',
            metavar="<name>",
            default=None,
            help=_('Filter results by name. Default=None.'),
        )
        parser.add_argument(
            '--description',
            metavar="<description>",
            default=None,
            help=_("Filter results by description. Default=None."),
        )
        parser.add_argument(
            "--name~",
            metavar="<name~>",
            default=None,
            help=_("Filter results matching a qos name pattern. "),
        )
        parser.add_argument(
            '--description~',
            metavar="<description~>",
            default=None,
            help=_("Filter results matching a qos description pattern."),
        )
        parser.add_argument(
            "--limit",
            metavar="<num-qos-types>",
            type=int,
            default=None,
            action=parseractions.NonNegativeAction,
            help=_("Limit the number of qos types returned. Default=None."),
        )
        parser.add_argument(
            '--offset',
            metavar="<offset>",
            default=None,
            help=_('Start position of qos type records listing.'),
        )
        parser.add_argument(
            '--sort-key',
            '--sort_key',
            metavar='<sort_key>',
            type=str,
            default=None,
            help=_(
                'Key to be sorted with, available keys are '
                '%(keys)s. Default=None.'
            )
            % {'keys': constants.QOS_TYPE_SORT_KEY_VALUES},
        )
        parser.add_argument(
            '--sort-dir',
            '--sort_dir',
            metavar='<sort_dir>',
            type=str,
            default=None,
            help=_(
                'Sort direction, available values are '
                '%(dirs)s. OPTIONAL: Default=None.'
            )
            % {'dirs': constants.SORT_DIR_VALUES},
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        search_opts = {
            'name': parsed_args.name,
            'description': parsed_args.description,
            'limit': parsed_args.limit,
            'offset': parsed_args.offset,
        }
        search_opts['name~'] = getattr(parsed_args, 'name~')
        search_opts['description~'] = getattr(parsed_args, 'description~')

        qos_types = share_client.qos_types.list(
            search_opts=search_opts,
            sort_key=parsed_args.sort_key,
            sort_dir=parsed_args.sort_dir,
        )

        formatted_types = []
        for qos_type in qos_types:
            formatted_types.append(
                format_qos_type(qos_type, parsed_args.formatter)
            )

        values = (
            oscutils.get_dict_properties(s._info, ATTRIBUTES)
            for s in formatted_types
        )

        columns = utils.format_column_headers(ATTRIBUTES)

        return (columns, values)


class ShowQosType(command.ShowOne):
    """Show a qos type."""

    _description = _("Display qos type details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_type',
            metavar="<qos_type>",
            help=_("Qos type to display (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        qos_type = apiutils.find_resource(
            share_client.qos_types, parsed_args.qos_type
        )

        formatted_type = format_qos_type(qos_type, parsed_args.formatter)

        return (
            ATTRIBUTES,
            oscutils.get_dict_properties(formatted_type._info, ATTRIBUTES),
        )



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manilaclient import api_versions
from manilaclient import base
from manilaclient import exceptions

RESOURCES_PATH = '/share-networks'
RESOURCE_PATH = "/share-networks/%s"
RESOURCE_NAME = 'share_network'
RESOURCES_NAME = 'share_networks'
ACTION_PATH = RESOURCE_PATH + '/action'


class ShareNetwork(base.Resource):
    """Network info for Manila shares."""

    def __repr__(self):
        return f"<ShareNetwork: {self.id}>"

    def update(self, **kwargs):
        """Update this share network."""
        return self.manager.update(self, **kwargs)

    def delete(self):
        """Delete this share network."""
        self.manager.delete(self)


class ShareNetworkManager(base.ManagerWithFind):
    """Manage :class:`ShareNetwork` resources."""

    resource_class = ShareNetwork

    @api_versions.wraps("1.0", "2.25")
    def create(
        self,
        neutron_net_id=None,
        neutron_subnet_id=None,
        nova_net_id=None,
        name=None,
        description=None,
    ):
        """Create share network.

        :param neutron_net_id: ID of Neutron network
        :param neutron_subnet_id: ID of Neutron subnet
        :param nova_net_id: ID of Nova network
        :param name: share network name
        :param description: share network description
        :rtype: :class:`ShareNetwork`
        """
        values = {}
        if neutron_net_id:
            values['neutron_net_id'] = neutron_net_id
        if neutron_subnet_id:
            values['neutron_subnet_id'] = neutron_subnet_id
        if nova_net_id:
            values['nova_net_id'] = nova_net_id
        if name:
            values['name'] = name
        if description:
            values['description'] = description

        body = {RESOURCE_NAME: values}

        return self._create(RESOURCES_PATH, body, RESOURCE_NAME)

    @api_versions.wraps("2.26", "2.50")  # noqa
    def create(  # noqa
        self,
        neutron_net_id=None,
        neutron_subnet_id=None,
        name=None,
        description=None,
    ):
        """Create share network.

        :param neutron_net_id: ID of Neutron network
        :param neutron_subnet_id: ID of Neutron subnet
        :param name: share network name
        :param description: share network description
        :rtype: :class:`ShareNetwork`
        """
        values = {}
        if neutron_net_id:
            values['neutron_net_id'] = neutron_net_id
        if neutron_subnet_id:
            values['neutron_subnet_id'] = neutron_subnet_id
        if name:
            values['name'] = name
        if description:
            values['description'] = description

        body = {RESOURCE_NAME: values}

        return self._create(RESOURCES_PATH, body, RESOURCE_NAME)

    @api_versions.wraps("2.51")  # noqa
    def create(  # noqa
        self,
        neutron_net_id=None,
        neutron_subnet_id=None,
        name=None,
        description=None,
        availability_zone=None,
    ):
        values = {}

        if neutron_net_id:
            values['neutron_net_id'] = neutron_net_id
        if neutron_subnet_id:
            values['neutron_subnet_id'] = neutron_subnet_id
        if name:
            values['name'] = name
        if description:
            values['description'] = description
        if availability_zone:
            values['availability_zone'] = availability_zone

        body = {RESOURCE_NAME: values}

        return self._create(RESOURCES_PATH, body, RESOURCE_NAME)

    def remove_security_service(self, share_network, security_service):
        """Dissociate security service from a share network.

        :param share_network: share network name, id or ShareNetwork instance
        :param security_service: name, id or SecurityService instance
        :rtype: :class:`ShareNetwork`
        """
        body = {
            'remove_security_service': {
                'security_service_id': base.getid(security_service),
            },
        }
        return self._create(
            RESOURCE_PATH % base.getid(share_network) + '/action',
            body,
            RESOURCE_NAME,
        )

    def get(self, share_network):
        """Get a share network.

        :param share_network: share network to get.
        :rtype: :class:`ShareNetwork`
        """
        return self._get(
            RESOURCE_PATH % base.getid(share_network), RESOURCE_NAME
        )

    @api_versions.wraps("1.0", "2.25")
    def update(
        self,
        share_network,
        neutron_net_id=None,
        neutron_subnet_id=None,
        nova_net_id=None,
        name=None,
        description=None,
    ):
        """Updates a share network.

        :param share_network: share network to update.
        :rtype: :class:`ShareNetwork`
        """
        values = {}
        if neutron_net_id is not None:
            values['neutron_net_id'] = neutron_net_id
        if neutron_subnet_id is not None:
            values['neutron_subnet_id'] = neutron_subnet_id
        if nova_net_id is not None:
            values['nova_net_id'] = nova_net_id
        if name is not None:
            values['name'] = name
        if description is not None:
            values['description'] = description

        for k, v in values.items():
            if v == '':
                values[k] = None

        if not values:
            msg = "Must specify fields to be updated"
            raise exceptions.CommandError(msg)

        body = {RESOURCE_NAME: values}
        return self._update(
            RESOURCE_PATH % base.getid(share_network), body, RESOURCE_NAME
        )

    @api_versions.wraps("2.26")  # noqa
    def update(  # noqa
        self,
        share_network,
        neutron_net_id=None,
        neutron_subnet_id=None,
        name=None,
        description=None,
    ):
        """Updates a share network.

        :param share_network: share network to update.
        :rtype: :class:`ShareNetwork`
        """
        values = {}
        if neutron_net_id is not None:
            values['neutron_net_id'] = neutron_net_id
        if neutron_subnet_id is not None:
            values['neutron_subnet_id'] = neutron_subnet_id
        if name is not None:
            values['name'] = name
        if description is not None:
            values['description'] = description

        for k, v in values.items():
            if v == '':
                values[k] = None

        if not values:
            msg = "Must specify fields to be updated"
            raise exceptions.CommandError(msg)

        body = {RESOURCE_NAME: values}
        return self._update(
            RESOURCE_PATH % base.getid(share_network), body, RESOURCE_NAME
        )

    def delete(self, share_network):
        """Delete a share network.

        :param share_network: share network to be deleted.
        """
        self._delete(RESOURCE_PATH % base.getid(share_network))

    def list(self, detailed=True, search_opts=None):
        """Get a list of all share network.

        :rtype: list of :class:`ShareNetwork`
        """
        query_string = self._build_query_string(search_opts)

        if detailed:
            path = RESOURCES_PATH + "/detail" + query_string
        else:
            path = RESOURCES_PATH + query_string

        return self._list(path, RESOURCES_NAME)

    def _action(self, action, share_network, info=None):
        """Perform a share network 'action'.

        :param action: text with action name.
        :param share_network: either share_network object or text with its ID.
        :param info: dict with data for specified 'action'.
        """
        body = {action: info}
        self.run_hooks('modify_body_for_action', body)
        url = ACTION_PATH % base.getid(share_network)
        return self.api.client.post(url, body=body)

    def add_security_service(self, share_network, security_service):
        """Associate given security service with a share network.

        :param share_network: share network name, id or ShareNetwork instance
        :param security_service: name, id or SecurityService instance
        :rtype: :class:`ShareNetwork`
        """
        info = {
            'security_service_id': base.getid(security_service),
        }
        return self._action('add_security_service', share_network, info)

    @api_versions.wraps("2.63")
    def add_security_service_check(
        self, share_network, security_service, reset_operation=False
    ):
        """Associate given security service with a share network.

        :param share_network: share network name, id or ShareNetwork instance
        :param security_service: name, id or SecurityService instance
        :param reset_operation: start over the check operation
        :rtype: :class:`ShareNetwork`
        """
        info = {
            'security_service_id': base.getid(security_service),
            'reset_operation': reset_operation,
        }
        return self._action('add_security_service_check', share_network, info)

    @api_versions.wraps("2.63")
    def update_share_network_security_service(
        self, share_network, current_security_service, new_security_service
    ):
        """Update current security service to new one of a given share network.

        :param share_network: share network name, id or ShareNetwork instance
        :param current_security_service: current name, id or
        SecurityService instance that will be changed
        :param new_security_service: new name, id or
        SecurityService instance that will be updated
        :rtype: :class:`ShareNetwork`
        """
        info = {
            'current_service_id': base.getid(current_security_service),
            'new_service_id': base.getid(new_security_service),
        }

        return self._action('update_security_service', share_network, info)

    @api_versions.wraps("2.63")
    def update_share_network_security_service_check(
        self,
        share_network,
        current_security_service,
        new_security_service,
        reset_operation=False,
    ):
        """Validates if the security service update is supported by all hosts.

        :param share_network: share network name, id or ShareNetwork instance
        :param current_security_service: current name, id or
        SecurityService instance that will be changed
        :param new_security_service: new name, id or
        :param reset_operation: start over the check operation
        SecurityService instance that will be updated
        :rtype: :class:`ShareNetwork`
        """
        info = {
            'current_service_id': base.getid(current_security_service),
            'new_service_id': base.getid(new_security_service),
            'reset_operation': reset_operation,
        }

        return self._action(
            'update_security_service_check', share_network, info
        )

    @api_versions.wraps("2.63")
    def reset_state(self, share_network, state):
        """Reset state of a share network.

        :param share_network: either share_network object or text with its ID
        or name.
        :param state: text with new state to set for share network.
        """
        return self._action('reset_status', share_network, {"status": state})

    @api_versions.wraps("2.70")
    def share_network_subnet_create_check(
        self,
        share_network_id,
        neutron_net_id=None,
        neutron_subnet_id=None,
        availability_zone=None,
        reset_operation=False,
    ):
        """Check if share network subnet can be created in given share network.

        :param share_network_id: ID of the Share network
        :param neutron_net_id: ID of Neutron network
        :param neutron_subnet_id: ID of Neutron subnet
        :param availability_zone: Name of the target availability zone
        :param reset_operation: Start over the check operation
        :rtype: :class:`ShareNetworkSubnet`
        """
        values = {}
        if neutron_net_id:
            values['neutron_net_id'] = neutron_net_id
        if neutron_subnet_id:
            values['neutron_subnet_id'] = neutron_subnet_id
        if availability_zone:
            values['availability_zone'] = availability_zone
        values['reset_operation'] = reset_operation

        return self._action(
            'share_network_subnet_create_check', share_network_id, values
        )



# Copyright (c) 2022 China Telecom Digital Intelligence.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manilaclient import api_versions
from manilaclient.tests.unit import utils
from manilaclient.tests.unit.v2 import fakes

TRANSFER_URL = 'share-transfers'
cs = fakes.FakeClient(api_versions.APIVersion('2.77'))


class ShareTransfersTest(utils.TestCase):
    def test_create(self):
        cs.transfers.create('1234')
        cs.assert_called(
            'POST',
            f'/{TRANSFER_URL}',
            body={'transfer': {'share_id': '1234', 'name': None}},
        )

    def test_get(self):
        transfer_id = '5678'
        cs.transfers.get(transfer_id)
        cs.assert_called('GET', f'/{TRANSFER_URL}/{transfer_id}')

    def test_list(self):
        cs.transfers.list()
        cs.assert_called('GET', f'/{TRANSFER_URL}/detail')

    def test_delete(self):
        cs.transfers.delete('5678')
        cs.assert_called('DELETE', f'/{TRANSFER_URL}/5678')

    def test_accept(self):
        transfer_id = '5678'
        auth_key = '12345'
        cs.transfers.accept(transfer_id, auth_key)
        cs.assert_called('POST', f'/{TRANSFER_URL}/{transfer_id}/accept')



#  Copyright (c) 2025 Cloudification GmbH.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from osc_lib import exceptions
from osc_lib import utils as oscutils

from manilaclient import api_versions
from manilaclient.common.apiclient.exceptions import BadRequest
from manilaclient.common.apiclient.exceptions import NotFound
from manilaclient.osc import utils
from manilaclient.osc.v2 import qos_types as osc_qos_types
from manilaclient.tests.unit.osc import osc_utils
from manilaclient.tests.unit.osc.v2 import fakes as manila_fakes

COLUMNS = [
    'id',
    'name',
    'description',
    'specs',
    'created_at',
    'updated_at',
]


class TestQosType(manila_fakes.TestShare):
    def setUp(self):
        super().setUp()

        self.qos_types_mock = self.app.client_manager.share.qos_types
        self.qos_types_mock.reset_mock()
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            api_versions.MAX_VERSION
        )


class TestQosTypeCreate(TestQosType):
    def setUp(self):
        super().setUp()

        self.new_qos_type = manila_fakes.FakeQosType.create_one_qostype()
        self.qos_types_mock.create.return_value = self.new_qos_type

        # Get the command object to test
        self.cmd = osc_qos_types.CreateQosType(self.app, None)

        self.data = [
            self.new_qos_type.id,
            self.new_qos_type.name,
            self.new_qos_type.description,
            ('expected_iops : 2000\npeak_iops : 5000'),
            self.new_qos_type.created_at,
            self.new_qos_type.updated_at,
        ]

        self.raw_data = [
            self.new_qos_type.id,
            self.new_qos_type.name,
            self.new_qos_type.description,
            {
                'expected_iops': '2000',
                'peak_iops': '5000',
            },
            self.new_qos_type.created_at,
            self.new_qos_type.updated_at,
        ]

    def test_qos_type_create_required_args(self):
        """Verifies required arguments."""

        arglist = [self.new_qos_type.name]
        verifylist = [
            ('name', self.new_qos_type.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.qos_types_mock.create.assert_called_with(
            name=self.new_qos_type.name,
        )

        self.assertCountEqual(COLUMNS, columns)
        self.assertCountEqual(self.data, data)

    def test_qos_type_create_json_format(self):
        """Verifies --format json."""

        arglist = [self.new_qos_type.name, '-f', 'json']
        verifylist = [
            ('name', self.new_qos_type.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.qos_types_mock.create.assert_called_with(
            name=self.new_qos_type.name,
        )

        self.assertCountEqual(COLUMNS, columns)
        self.assertCountEqual(self.raw_data, data)

    def test_qos_type_create_missing_required_arg(self):
        """Verifies missing required arguments."""

        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_qos_type_create_specs(self):
        arglist = [
            self.new_qos_type.name,
            '--spec',
            'peak_iops=100',
            '--spec',
            'expected_iops=20',
        ]
        verifylist = [
            ('name', self.new_qos_type.name),
            ('spec', ['peak_iops=100', 'expected_iops=20']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.qos_types_mock.create.assert_called_with(
            name=self.new_qos_type.name,
            specs={'peak_iops': '100', 'expected_iops': '20'},
        )

        self.assertCountEqual(COLUMNS, columns)
        self.assertCountEqual(self.data, data)


class TestQosTypeDelete(TestQosType):
    qos_types = manila_fakes.FakeQosType.create_qos_types(count=2)

    def setUp(self):
        super().setUp()

        self.qos_types_mock.get = manila_fakes.FakeQosType.get_qos_types(
            self.qos_types
        )
        self.qos_types_mock.delete.return_value = None

        # Get the command object to test
        self.cmd = osc_qos_types.DeleteQosType(self.app, None)

    def test_qos_type_delete_one(self):
        arglist = [self.qos_types[0].id]

        verifylist = [('qos_types', [self.qos_types[0].id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.qos_types_mock.delete.assert_called_with(self.qos_types[0])
        self.assertIsNone(result)

    def test_qos_type_delete_multiple(self):
        arglist = []
        for t in self.qos_types:
            arglist.append(t.id)
        verifylist = [
            ('qos_types', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        calls = []
        for t in self.qos_types:
            calls.append(mock.call(t))
        self.qos_types_mock.delete.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_delete_qos_type_with_exception(self):
        arglist = [
            'non_existing_type',
        ]
        verifylist = [
            ('qos_types', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.qos_types_mock.delete.side_effect = exceptions.CommandError()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestQosTypeSet(TestQosType):
    def setUp(self):
        super().setUp()

        self.qos_type = manila_fakes.FakeQosType.create_one_qostype(
            methods={'set_keys': None, 'update': None}
        )
        self.qos_types_mock.get.return_value = self.qos_type

        # Get the command object to test
        self.cmd = osc_qos_types.SetQosType(self.app, None)

    def test_qos_type_set_specs(self):
        arglist = [
            self.qos_type.id,
            '--spec',
            'peak_iops=100',
        ]
        verifylist = [
            ('qos_type', self.qos_type.id),
            ('spec', ['peak_iops=100']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.qos_type.set_keys.assert_called_with({'peak_iops': '100'})
        self.assertIsNone(result)

    def test_qos_type_set_description(self):
        arglist = [self.qos_type.id, '--description', 'new description']
        verifylist = [
            ('qos_type', self.qos_type.id),
            ('description', 'new description'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.qos_type.update.assert_called_with(description='new description')
        self.assertIsNone(result)

    def test_qos_type_set_specs_exception(self):
        arglist = [
            self.qos_type.id,
            '--spec',
            'peak_iops=100',
        ]
        verifylist = [
            ('qos_type', self.qos_type.id),
            ('spec', ['peak_iops=100']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.qos_type.set_keys.side_effect = BadRequest()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestQosTypeUnset(TestQosType):
    def setUp(self):
        super().setUp()

        self.qos_type = manila_fakes.FakeQosType.create_one_qostype(
            methods={'unset_keys': None, 'update': None}
        )

        self.qos_types_mock.get.return_value = self.qos_type

        # Get the command object to test
        self.cmd = osc_qos_types.UnsetQosType(self.app, None)

    def test_qos_type_unset_description(self):
        arglist = [self.qos_type.id, '--description']
        verifylist = [
            ('qos_type', self.qos_type.id),
            ('description', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.qos_type.update.assert_called_with(description=None)
        self.assertIsNone(result)

    def test_qos_type_unset_specs(self):
        arglist = [self.qos_type.id, '--spec', 'peak_iops']
        verifylist = [
            ('qos_type', self.qos_type.id),
            ('spec', ['peak_iops']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.qos_type.unset_keys.assert_called_with(['peak_iops'])
        self.assertIsNone(result)

    def test_qos_type_unset_exception(self):
        arglist = [self.qos_type.id, '--spec', 'peak_iops']
        verifylist = [
            ('qos_type', self.qos_type.id),
            ('spec', ['peak_iops']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.qos_type.unset_keys.side_effect = NotFound()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestQosTypeList(TestQosType):
    qos_types = manila_fakes.FakeQosType.create_qos_types()
    columns = utils.format_column_headers(COLUMNS)

    def setUp(self):
        super().setUp()

        self.qos_types_mock.list.return_value = self.qos_types

        # Get the command object to test
        self.cmd = osc_qos_types.ListQosType(self.app, None)

        self.values = (
            oscutils.get_dict_properties(s._info, COLUMNS)
            for s in self.qos_types
        )

    def test_qos_type_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.qos_types_mock.list.assert_called_once_with(
            search_opts={
                'offset': None,
                'limit': None,
                'name': None,
                'description': None,
                'name~': None,
                'description~': None,
            },
            sort_key=None,
            sort_dir=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.values), list(data))

    def test_qos_type_list_by_name(self):
        arglist = ['--name', 'fake_name']
        verifylist = [('name', 'fake_name')]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.qos_types_mock.list.assert_called_with(
            search_opts={
                'offset': None,
                'limit': None,
                'name': 'fake_name',
                'description': None,
                'name~': None,
                'description~': None,
            },
            sort_key=None,
            sort_dir=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(list(self.values), list(data))


class TestQosTypeShow(TestQosType):
    def setUp(self):
        super().setUp()

        self.qos_type = manila_fakes.FakeQosType.create_one_qostype()

        self.qos_types_mock.get.return_value = self.qos_type

        # Get the command object to test
        self.cmd = osc_qos_types.ShowQosType(self.app, None)

        self.data = [
            self.qos_type.id,
            self.qos_type.name,
            self.qos_type.description,
            ('expected_iops : 2000\npeak_iops : 5000'),
            self.qos_type.created_at,
            self.qos_type.updated_at,
        ]

        self.raw_data = [
            self.qos_type.id,
            self.qos_type.name,
            self.qos_type.description,
            {
                'expected_iops': '2000',
                'peak_iops': '5000',
            },
            self.qos_type.created_at,
            self.qos_type.updated_at,
        ]

    def test_qos_type_show(self):
        arglist = [self.qos_type.id]
        verifylist = [("qos_type", self.qos_type.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.qos_types_mock.get.assert_called_with(self.qos_type.id)

        self.assertCountEqual(COLUMNS, columns)
        self.assertCountEqual(self.data, data)

    def test_qos_type_show_json_format(self):
        arglist = [
            self.qos_type.id,
            '-f',
            'json',
        ]
        verifylist = [("qos_type", self.qos_type.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.qos_types_mock.get.assert_called_with(self.qos_type.id)

        self.assertCountEqual(COLUMNS, columns)
        self.assertCountEqual(self.raw_data, data)



# Copyright 2019 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.


"""OpenStackClient plugin for the Shared File System Service."""

import logging

from osc_lib import utils

from manilaclient import api_versions
from manilaclient import client
from manilaclient.common import constants
from manilaclient import exceptions

LOG = logging.getLogger(__name__)

API_NAME = 'share'
API_VERSION_OPTION = 'os_share_api_version'
CLIENT_CLASS = 'manilaclient.v2.client.Client'
LATEST_VERSION = api_versions.MAX_VERSION
LATEST_MINOR_VERSION = api_versions.MAX_VERSION.split('.')[-1]

API_VERSIONS = {
    f'2.{i}': CLIENT_CLASS for i in range(0, int(LATEST_MINOR_VERSION) + 1)
}


def _get_manila_url_from_service_catalog(instance):
    service_type = constants.SFS_SERVICE_TYPE
    url = instance.get_endpoint_for_service_type(
        constants.SFS_SERVICE_TYPE,
        region_name=instance._region_name,
        interface=instance.interface,
    )
    # Fallback if cloud is using an older service type name
    if not url:
        url = instance.get_endpoint_for_service_type(
            constants.V2_SERVICE_TYPE,
            region_name=instance._region_name,
            interface=instance.interface,
        )
        service_type = constants.V2_SERVICE_TYPE
    if url is None:
        raise exceptions.EndpointNotFound(
            message="Could not find manila / shared-file-system endpoint in "
            "the service catalog."
        )
    return service_type, url


def make_client(instance):
    """Returns a shared file system service client."""
    requested_api_version = instance._api_version[API_NAME]

    endpoint_override = getattr(
        instance._cli_options, 'os_endpoint_override', None
    )
    if endpoint_override:
        manila_endpoint_url = endpoint_override
        service_type = constants.SFS_SERVICE_TYPE
    else:
        service_type, manila_endpoint_url = (
            _get_manila_url_from_service_catalog(instance)
        )

    instance.setup_auth()
    debugging_enabled = instance._cli_options.debug

    client_args = dict(
        session=instance.session,
        service_catalog_url=manila_endpoint_url,
        endpoint_type=instance.interface,
        region_name=instance.region_name,
        service_type=service_type,
        auth=instance.auth,
        http_log_debug=debugging_enabled,
        cacert=instance.cacert,
        cert=instance.cert,
        insecure=not instance.verify,
    )

    # Cast the API version into an object for further processing
    requested_api_version = api_versions.APIVersion(
        version_str=requested_api_version
    )

    max_version = api_versions.APIVersion(api_versions.MAX_VERSION)
    client_args.update(dict(api_version=max_version))
    temp_client = client.Client(max_version, **client_args)
    discovered_version = api_versions.discover_version(
        temp_client, requested_api_version
    )

    shared_file_system_client = utils.get_client_class(
        API_NAME, discovered_version.get_string(), API_VERSIONS
    )

    LOG.debug(
        'Instantiating Shared File System (share) client: %s',
        shared_file_system_client,
    )
    LOG.debug('Shared File System API version: %s', discovered_version)

    client_args.update(dict(api_version=discovered_version))
    return shared_file_system_client(**client_args)


def build_option_parser(parser):
    """Hook to add global options."""
    default_api_version = utils.env('OS_SHARE_API_VERSION') or LATEST_VERSION
    parser.add_argument(
        '--os-share-api-version',
        metavar='<shared-file-system-api-version>',
        default=default_api_version,
        choices=sorted(
            API_VERSIONS, key=lambda k: [int(x) for x in k.split('.')]
        ),
        help='Shared File System API version, default='
        + default_api_version
        + ' (version supported by both the client and the server) '
        '(Env: OS_SHARE_API_VERSION)',
    )
    parser.add_argument(
        "--os-endpoint-override",
        metavar="<endpoint-override>",
        default=utils.env(
            "OS_SHARED_FILE_SYSTEM_ENDPOINT_OVERRIDE",
            "OS_ENDPOINT_OVERRIDE",
            "OS_MANILA_BYPASS_URL",
            "MANILACLIENT_BYPASS_URL",
        ),
        help=(
            "Use this API endpoint instead of the Service Catalog. "
            "Defaults to env[OS_SHARED_FILE_SYSTEM_ENDPOINT_OVERRIDE]."
        ),
    )
    return parser



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

import ddt
from osc_lib import exceptions
from osc_lib import utils as oscutils

from manilaclient import api_versions
from manilaclient.osc.v2 import share_access_rules as osc_share_access_rules
from manilaclient.tests.unit.osc.v2 import fakes as manila_fakes

ACCESS_RULE_ATTRIBUTES = [
    'id',
    'share_id',
    'access_level',
    'access_to',
    'access_type',
    'state',
    'access_key',
    'created_at',
    'updated_at',
    'properties',
]


class TestShareAccess(manila_fakes.TestShare):
    def setUp(self):
        super().setUp()

        self.shares_mock = self.app.client_manager.share.shares
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            api_versions.MAX_VERSION
        )
        self.shares_mock.reset_mock()

        self.access_rules_mock = (
            self.app.client_manager.share.share_access_rules
        )
        self.access_rules_mock.reset_mock()


@ddt.ddt
class TestShareAccessCreate(TestShareAccess):
    def setUp(self):
        super().setUp()

        self.share = manila_fakes.FakeShare.create_one_share(
            attrs={"is_public": False}, methods={'allow': None}
        )
        self.shares_mock.get.return_value = self.share
        self.access_rule = (
            manila_fakes.FakeShareAccessRule.create_one_access_rule(
                attrs={"share_id": self.share.id}
            )
        )
        self.share.allow.return_value = self.access_rule._info
        self.access_rules_mock.get.return_value = self.access_rule

        # Get the command object to test
        self.cmd = osc_share_access_rules.ShareAccessAllow(self.app, None)

    def test_share_access_create_user(self):
        arglist = [
            self.share.id,
            'user',
            'demo',
        ]
        verifylist = [
            ("share", self.share.id),
            ("access_type", "user"),
            ("access_to", "demo"),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.shares_mock.get.assert_called_with(self.share.id)
        self.share.allow.assert_called_with(
            access_type="user",
            access="demo",
            access_level=None,
            metadata={},
        )
        self.assertEqual(ACCESS_RULE_ATTRIBUTES, columns)
        self.assertCountEqual(self.access_rule._info.values(), data)

    def test_share_access_create_properties(self):
        arglist = [self.share.id, 'user', 'demo', '--properties', 'key=value']
        verifylist = [
            ("share", self.share.id),
            ("access_type", "user"),
            ("access_to", "demo"),
            ('properties', ['key=value']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.shares_mock.get.assert_called_with(self.share.id)
        self.share.allow.assert_called_with(
            access_type="user",
            access="demo",
            access_level=None,
            metadata={'key': 'value'},
        )
        self.assertEqual(ACCESS_RULE_ATTRIBUTES, columns)
        self.assertCountEqual(self.access_rule._info.values(), data)

    @ddt.data(
        {
            'lock_visibility': True,
            'lock_deletion': True,
            'lock_reason': 'testing resource locks',
        },
        {'lock_visibility': False, 'lock_deletion': True, 'lock_reason': None},
        {'lock_visibility': True, 'lock_deletion': False, 'lock_reason': None},
    )
    @ddt.unpack
    def test_share_access_create_restrict(
        self, lock_visibility, lock_deletion, lock_reason
    ):
        arglist = [self.share.id, 'user', 'demo', '--properties', 'key=value']
        verifylist = [
            ("share", self.share.id),
            ("access_type", "user"),
            ("access_to", "demo"),
            ('properties', ['key=value']),
        ]
        allow_call_kwargs = {}
        if lock_visibility:
            arglist.append('--lock-visibility')
            verifylist.append(('lock_visibility', lock_visibility))
            allow_call_kwargs['lock_visibility'] = lock_visibility
        if lock_deletion:
            arglist.append('--lock-deletion')
            verifylist.append(('lock_deletion', lock_deletion))
            allow_call_kwargs['lock_deletion'] = lock_deletion
        if lock_reason:
            arglist.append('--lock-reason')
            arglist.append(lock_reason)
            verifylist.append(('lock_reason', lock_reason))
            allow_call_kwargs['lock_reason'] = lock_reason

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.shares_mock.get.assert_called_with(self.share.id)
        self.share.allow.assert_called_with(
            access_type="user",
            access="demo",
            access_level=None,
            metadata={'key': 'value'},
            **allow_call_kwargs,
        )
        self.assertEqual(ACCESS_RULE_ATTRIBUTES, columns)
        self.assertCountEqual(self.access_rule._info.values(), data)

    @ddt.data(
        {'lock_visibility': True, 'lock_deletion': False},
        {'lock_visibility': False, 'lock_deletion': True},
    )
    @ddt.unpack
    def test_share_access_create_restrict_not_available(
        self, lock_visibility, lock_deletion
    ):
        arglist = [
            self.share.id,
            'user',
            'demo',
        ]
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            "2.79"
        )
        verifylist = [
            ("share", self.share.id),
            ("access_type", "user"),
            ("access_to", "demo"),
            ("lock_visibility", lock_visibility),
            ("lock_deletion", lock_deletion),
            ("lock_reason", None),
        ]
        if lock_visibility:
            arglist.append('--lock-visibility')
        if lock_deletion:
            arglist.append('--lock-deletion')
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_access_rule_create_access_level(self):
        arglist = [self.share.id, 'user', 'demo', '--access-level', 'ro']
        verifylist = [
            ("share", self.share.id),
            ("access_type", "user"),
            ("access_to", "demo"),
            ('access_level', 'ro'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.shares_mock.get.assert_called_with(self.share.id)
        self.share.allow.assert_called_with(
            access_type="user",
            access="demo",
            access_level='ro',
            metadata={},
        )
        self.assertEqual(ACCESS_RULE_ATTRIBUTES, columns)
        self.assertCountEqual(self.access_rule._info.values(), data)

    def test_share_access_create_wait(self):
        arglist = [self.share.id, 'user', 'demo', '--wait']
        verifylist = [
            ("share", self.share.id),
            ("access_type", "user"),
            ("access_to", "demo"),
            ("wait", True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.shares_mock.get.assert_called_with(self.share.id)
        self.share.allow.assert_called_with(
            access_type="user",
            access="demo",
            access_level=None,
            metadata={},
        )
        self.assertEqual(ACCESS_RULE_ATTRIBUTES, columns)
        self.assertCountEqual(self.access_rule._info.values(), data)

    @mock.patch('manilaclient.osc.v2.share_access_rules.LOG')
    def test_share_access_create_wait_error(self, mock_logger):
        arglist = [self.share.id, 'user', 'demo', '--wait']
        verifylist = [
            ("share", self.share.id),
            ("access_type", "user"),
            ("access_to", "demo"),
            ("wait", True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_status', return_value=False):
            columns, data = self.cmd.take_action(parsed_args)

            self.shares_mock.get.assert_called_with(self.share.id)
            self.share.allow.assert_called_with(
                access_type="user",
                access="demo",
                access_level=None,
                metadata={},
            )

            mock_logger.error.assert_called_with(
                "ERROR: Share access rule is in error state."
            )

            self.assertEqual(ACCESS_RULE_ATTRIBUTES, columns)
            self.assertCountEqual(self.access_rule._info.values(), data)


@ddt.ddt
class TestShareAccessDelete(TestShareAccess):
    def setUp(self):
        super().setUp()

        self.share = manila_fakes.FakeShare.create_one_share(
            methods={'deny': None}
        )
        self.shares_mock.get.return_value = self.share
        self.access_rule = (
            manila_fakes.FakeShareAccessRule.create_one_access_rule(
                attrs={"share_id": self.share.id}
            )
        )

        # Get the command object to test
        self.cmd = osc_share_access_rules.ShareAccessDeny(self.app, None)

    @ddt.data(True, False)
    def test_share_access_delete(self, unrestrict):
        arglist = [self.share.id, self.access_rule.id]
        verifylist = [
            ("share", self.share.id),
            ("id", self.access_rule.id),
        ]
        deny_kwargs = {}
        if unrestrict:
            arglist.append('--unrestrict')
            verifylist.append(("unrestrict", unrestrict))
            deny_kwargs['unrestrict'] = unrestrict
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.shares_mock.get.assert_called_with(self.share.id)
        self.share.deny.assert_called_with(self.access_rule.id, **deny_kwargs)
        self.assertIsNone(result)

    def test_share_access_delete_unrestrict_not_available(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            "2.79"
        )
        arglist = [self.share.id, self.access_rule.id, "--unrestrict"]
        verifylist = [
            ("share", self.share.id),
            ("id", self.access_rule.id),
            ("unrestrict", True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_access_delete_wait(self):
        arglist = [self.share.id, self.access_rule.id, '--wait']
        verifylist = [
            ("share", self.share.id),
            ("id", self.access_rule.id),
            ('wait', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_delete', return_value=True):
            result = self.cmd.take_action(parsed_args)

            self.shares_mock.get.assert_called_with(self.share.id)
            self.share.deny.assert_called_with(self.access_rule.id)
            self.assertIsNone(result)

    def test_share_access_delete_wait_error(self):
        arglist = [self.share.id, self.access_rule.id, '--wait']
        verifylist = [
            ("share", self.share.id),
            ("id", self.access_rule.id),
            ('wait', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        with mock.patch('osc_lib.utils.wait_for_delete', return_value=False):
            self.assertRaises(
                exceptions.CommandError, self.cmd.take_action, parsed_args
            )


@ddt.ddt
class TestShareAccessList(TestShareAccess):
    access_rules_columns = [
        'ID',
        'Access Type',
        'Access To',
        'Access Level',
        'State',
        'Access Key',
        'Created At',
        'Updated At',
    ]

    def setUp(self):
        super().setUp()

        self.share = manila_fakes.FakeShare.create_one_share()
        self.access_rule_1 = (
            manila_fakes.FakeShareAccessRule.create_one_access_rule(
                attrs={"share_id": self.share.id}
            )
        )
        self.access_rule_2 = (
            manila_fakes.FakeShareAccessRule.create_one_access_rule(
                attrs={"share_id": self.share.id, "access_to": "admin"}
            )
        )
        self.access_rules = [self.access_rule_1, self.access_rule_2]

        self.shares_mock.get.return_value = self.share
        self.access_rules_mock.access_list.return_value = self.access_rules
        self.values_list = (
            oscutils.get_dict_properties(a._info, self.access_rules_columns)
            for a in self.access_rules
        )

        # Get the command object to test
        self.cmd = osc_share_access_rules.ListShareAccess(self.app, None)

    def test_access_rules_list(self):
        arglist = [self.share.id]
        verifylist = [("share", self.share.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.shares_mock.get.assert_called_with(self.share.id)
        self.access_rules_mock.access_list.assert_called_with(self.share, {})
        self.assertEqual(self.access_rules_columns, columns)
        self.assertEqual(tuple(self.values_list), tuple(data))

    def test_access_rules_list_filter_properties(self):
        arglist = [self.share.id, '--properties', 'key=value']
        verifylist = [("share", self.share.id), ('properties', ['key=value'])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.shares_mock.get.assert_called_with(self.share.id)
        self.access_rules_mock.access_list.assert_called_with(
            self.share, {'metadata': {'key': 'value'}}
        )
        self.assertEqual(self.access_rules_columns, columns)
        self.assertEqual(tuple(self.values_list), tuple(data))

    @ddt.data(
        {'access_to': '10.0.0.0/0', 'access_type': 'ip'},
        {'access_key': '10.0.0.0/0', 'access_level': 'rw'},
    )
    def test_access_rules_list_access_filters(self, filters):
        arglist = [
            self.share.id,
        ]

        verifylist = [
            ("share", self.share.id),
        ]
        for filter_key, filter_value in filters.items():
            filter_arg = filter_key.replace("_", "-")
            arglist.append(f'--{filter_arg}')
            arglist.append(filter_value)
            verifylist.append((filter_key, filter_value))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.shares_mock.get.assert_called_with(self.share.id)
        self.access_rules_mock.access_list.assert_called_with(
            self.share, filters
        )
        self.assertEqual(self.access_rules_columns, columns)
        self.assertEqual(tuple(self.values_list), tuple(data))

    @ddt.data(
        {'access_to': '10.0.0.0/0', 'access_type': 'ip'},
        {'access_key': '10.0.0.0/0', 'access_level': 'rw'},
    )
    def test_access_rules_list_access_filters_command_error(self, filters):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            "2.81"
        )
        arglist = [
            self.share.id,
        ]
        verifylist = [
            ("share", self.share.id),
        ]
        for filter_key, filter_value in filters.items():
            filter_arg = filter_key.replace("_", "-")
            arglist.append(f'--{filter_arg}')
            arglist.append(filter_value)
            verifylist.append((filter_key, filter_value))
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShareAccessShow(TestShareAccess):
    def setUp(self):
        super().setUp()

        self.share = manila_fakes.FakeShare.create_one_share()
        self.access_rule = (
            manila_fakes.FakeShareAccessRule.create_one_access_rule(
                attrs={"share_id": self.share.id}
            )
        )
        self.access_rules_mock.get.return_value = self.access_rule

        # Get the command object to test
        self.cmd = osc_share_access_rules.ShowShareAccess(self.app, None)

    def test_access_rule_show(self):
        arglist = [self.access_rule.id]
        verifylist = [("access_id", self.access_rule.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.access_rules_mock.get.assert_called_with(self.access_rule.id)
        self.assertEqual(ACCESS_RULE_ATTRIBUTES, columns)
        self.assertEqual(tuple(self.access_rule._info.values()), data)


class TestShareAccessSet(TestShareAccess):
    def setUp(self):
        super().setUp()

        self.share = manila_fakes.FakeShare.create_one_share()
        self.access_rule = (
            manila_fakes.FakeShareAccessRule.create_one_access_rule(
                attrs={"share_id": self.share.id}
            )
        )
        self.access_rules_mock.get.return_value = self.access_rule

        # Get the command object to test
        self.cmd = osc_share_access_rules.SetShareAccess(self.app, None)

    def test_access_rule_set_metadata(self):
        arglist = [self.access_rule.id, '--property', 'key1=value1']
        verifylist = [
            ("access_id", self.access_rule.id),
            ('property', {'key1': 'value1'}),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.access_rules_mock.set_metadata.assert_called_with(
            self.access_rule, {'key1': 'value1'}
        )
        self.assertIsNone(result)

    def test_access_rule_set_access_level(self):
        arglist = [self.access_rule.id, '--access-level', 'ro']
        verifylist = [
            ("access_id", self.access_rule.id),
            ('access_level', 'ro'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.access_rules_mock.set_access_level.assert_called_with(
            self.access_rule, 'ro'
        )
        self.assertIsNone(result)


class TestShareAccessUnset(TestShareAccess):
    def setUp(self):
        super().setUp()

        self.share = manila_fakes.FakeShare.create_one_share()
        self.access_rule = (
            manila_fakes.FakeShareAccessRule.create_one_access_rule(
                attrs={"share_id": self.share.id}
            )
        )
        self.access_rules_mock.get.return_value = self.access_rule

        # Get the command object to test
        self.cmd = osc_share_access_rules.UnsetShareAccess(self.app, None)

    def test_access_rule_unset(self):
        arglist = [self.access_rule.id, '--property', 'key1']
        verifylist = [
            ("access_id", self.access_rule.id),
            ('property', ['key1']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.access_rules_mock.unset_metadata.assert_called_with(
            self.access_rule, ['key1']
        )
        self.assertIsNone(result)



# Copyright (c) 2022 China Telecom Digital Intelligence.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manilaclient.tests.functional.osc import base


class TransfersCLITest(base.OSCClientTestBase):
    def setUp(self):
        super().setUp()
        self.share_type = self.create_share_type()

    def test_transfer_create_list_show_delete(self):
        share = self.create_share(
            share_type=self.share_type['name'],
            wait_for_status='available',
            client=self.user_client,
        )
        # create share transfer
        self.create_share_transfer(share['id'], name='transfer_test')
        self._wait_for_object_status('share', share['id'], 'awaiting_transfer')

        # Get all transfers
        transfers = self.listing_result(
            'share', 'transfer list', client=self.user_client
        )
        # We must have at least one transfer
        self.assertTrue(len(transfers) > 0)
        self.assertTableStruct(
            transfers,
            [
                'ID',
                'Name',
                'Resource Type',
                'Resource Id',
            ],
        )

        # grab the transfer we created
        transfer = [
            transfer
            for transfer in transfers
            if transfer['Resource Id'] == share['id']
        ]
        self.assertEqual(1, len(transfer))

        show_transfer = self.dict_result(
            'share', f'transfer show {transfer[0]["ID"]}'
        )
        self.assertEqual(transfer[0]['ID'], show_transfer['id'])
        expected_keys = (
            'id',
            'created_at',
            'name',
            'resource_type',
            'resource_id',
            'source_project_id',
            'destination_project_id',
            'accepted',
            'expires_at',
        )
        for key in expected_keys:
            self.assertIn(key, show_transfer)

        # filtering by Resource ID
        filtered_transfers = self.listing_result(
            'share',
            f'transfer list --resource-id {share["id"]}',
            client=self.user_client,
        )
        self.assertEqual(1, len(filtered_transfers))
        self.assertEqual(show_transfer['resource_id'], share["id"])

        # finally delete transfer and share
        self.openstack(f'share transfer delete {show_transfer["id"]}')
        self._wait_for_object_status('share', share['id'], 'available')

    def test_transfer_accept(self):
        share = self.create_share(
            share_type=self.share_type['name'],
            wait_for_status='available',
            client=self.user_client,
        )
        # create share transfer
        transfer = self.create_share_transfer(
            share['id'], name='transfer_test'
        )
        self._wait_for_object_status('share', share['id'], 'awaiting_transfer')

        # accept share transfer
        self.openstack(
            f'share transfer accept {transfer["id"]} {transfer["auth_key"]}'
        )
        self._wait_for_object_status('share', share['id'], 'available')



# Copyright (c) 2022 China Telecom Digital Intelligence.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manilaclient import api_versions
from manilaclient import base
from manilaclient.common import constants


class ShareTransfer(base.Resource):
    """Transfer a share from one tenant to another"""

    def __repr__(self):
        return f"<ShareTransfer: {self.id}>"

    def delete(self):
        """Delete this share transfer."""
        return self.manager.delete(self)


class ShareTransferManager(base.ManagerWithFind):
    """Manage :class:`ShareTransfer` resources."""

    resource_class = ShareTransfer

    @api_versions.wraps(constants.SHARE_TRANSFER_VERSION)
    def create(self, share_id, name=None):
        """Creates a share transfer.

        :param share_id: The ID of the share to transfer.
        :param name: The name of the transfer.
        :rtype: :class:`ShareTransfer`
        """
        body = {'transfer': {'share_id': share_id, 'name': name}}

        return self._create('/share-transfers', body, 'transfer')

    @api_versions.wraps(constants.SHARE_TRANSFER_VERSION)
    def accept(self, transfer, auth_key, clear_access_rules=False):
        """Accept a share transfer.

        :param transfer_id: The ID of the transfer to accept.
        :param auth_key: The auth_key of the transfer.
        :param clear_access_rules: Transfer share without access rules
        :rtype: :class:`ShareTransfer`
        """
        transfer_id = base.getid(transfer)
        body = {
            'accept': {
                'auth_key': auth_key,
                'clear_access_rules': clear_access_rules,
            }
        }

        self._accept(f'/share-transfers/{transfer_id}/accept', body)

    @api_versions.wraps(constants.SHARE_TRANSFER_VERSION)
    def get(self, transfer_id):
        """Show details of a share transfer.

        :param transfer_id: The ID of the share transfer to display.
        :rtype: :class:`ShareTransfer`
        """
        return self._get(f"/share-transfers/{transfer_id}", "transfer")

    @api_versions.wraps(constants.SHARE_TRANSFER_VERSION)
    def list(
        self, detailed=True, search_opts=None, sort_key=None, sort_dir=None
    ):
        """Get a list of all share transfer.

        :param detailed: Get detailed object information.
        :param search_opts: Filtering options.
        :param sort_key: Key to be sorted (i.e. 'created_at').
        :param sort_dir: Sort direction, should be 'desc' or 'asc'.
        :rtype: list of :class:`ShareTransfer`
        """
        if search_opts is None:
            search_opts = {}

        if sort_key is not None:
            if sort_key in constants.SHARE_TRANSFER_SORT_KEY_VALUES:
                search_opts['sort_key'] = sort_key
                # NOTE: Replace aliases with appropriate keys
                if sort_key == 'name':
                    search_opts['sort_key'] = 'display_name'
            else:
                raise ValueError(
                    'sort_key must be one of the following: {}.'.format(
                        ', '.join(constants.SHARE_TRANSFER_SORT_KEY_VALUES)
                    )
                )

        if sort_dir is not None:
            if sort_dir in constants.SORT_DIR_VALUES:
                search_opts['sort_dir'] = sort_dir
            else:
                raise ValueError(
                    'sort_dir must be one of the following: {}.'.format(
                        ', '.join(constants.SORT_DIR_VALUES)
                    )
                )

        query_string = self._build_query_string(search_opts)

        if detailed:
            path = f"/share-transfers/detail{query_string}"
        else:
            path = f"/share-transfers{query_string}"

        return self._list(path, 'transfers')

    @api_versions.wraps(constants.SHARE_TRANSFER_VERSION)
    def delete(self, transfer_id):
        """Delete a share transfer.

        :param transfer_id: The :class:`ShareTransfer` to delete.
        """

        return self._delete(f"/share-transfers/{base.getid(transfer_id)}")



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
from unittest import mock

from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils as oscutils

from manilaclient import api_versions
from manilaclient.common import cliutils

from manilaclient.osc import utils
from manilaclient.osc.v2 import share_replicas as osc_share_replicas

from manilaclient.tests.unit.osc import osc_utils
from manilaclient.tests.unit.osc.v2 import fakes as manila_fakes


class TestShareReplica(manila_fakes.TestShare):
    def setUp(self):
        super().setUp()

        self.shares_mock = self.app.client_manager.share.shares
        self.shares_mock.reset_mock()

        self.replicas_mock = self.app.client_manager.share.share_replicas
        self.replicas_mock.reset_mock()
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            api_versions.MAX_VERSION
        )

        self.replica_el_mock = (
            self.app.client_manager.share.share_replica_export_locations
        )
        self.replica_el_mock.reset_mock()


class TestShareReplicaCreate(TestShareReplica):
    def setUp(self):
        super().setUp()

        self.share = manila_fakes.FakeShare.create_one_share()
        self.shares_mock.get.return_value = self.share

        self.share_replica = manila_fakes.FakeShareReplica.create_one_replica(
            attrs={'availability_zone': 'manila-zone-1', 'status': 'available'}
        )
        self.replicas_mock.create.return_value = self.share_replica
        self.replicas_mock.get.return_value = self.share_replica

        self.cmd = osc_share_replicas.CreateShareReplica(self.app, None)

        self.data = tuple(self.share_replica._info.values())
        self.columns = tuple(self.share_replica._info.keys())

    def test_share_replica_create_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_replica_create(self):
        arglist = [self.share.id]
        verifylist = [('share', self.share.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.replicas_mock.create.assert_called_with(
            share=self.share, availability_zone=None
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_replica_create_az(self):
        arglist = [
            self.share.id,
            '--availability-zone',
            self.share.availability_zone,
        ]
        verifylist = [
            ('share', self.share.id),
            ('availability_zone', self.share.availability_zone),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.replicas_mock.create.assert_called_with(
            share=self.share, availability_zone=self.share.availability_zone
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_replica_create_scheduler_hint_valid(self):
        arglist = [
            self.share.id,
            '--availability-zone',
            self.share.availability_zone,
            '--scheduler-hint',
            ('only_host=host1@backend1#pool1'),
        ]
        verifylist = [
            ('share', self.share.id),
            ('availability_zone', self.share.availability_zone),
            ('scheduler_hint', {'only_host': 'host1@backend1#pool1'}),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.replicas_mock.create.assert_called_with(
            share=self.share,
            availability_zone=self.share.availability_zone,
            scheduler_hints={'only_host': 'host1@backend1#pool1'},
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_replica_create_scheduler_hint_invalid_hint(self):
        arglist = [
            self.share.id,
            '--availability-zone',
            self.share.availability_zone,
            '--scheduler-hint',
            'fake_hint=host1@backend1#pool1',
        ]
        verifylist = [
            ('share', self.share.id),
            ('availability_zone', self.share.availability_zone),
            ('scheduler_hint', {'fake_hint': 'host1@backend1#pool1'}),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_replica_create_scheduler_hint_invalid_version(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            "2.66"
        )

        arglist = [
            self.share.id,
            '--availability-zone',
            self.share.availability_zone,
            '--scheduler-hint',
            'only_host=host1@backend1#pool1',
        ]
        verifylist = [
            ('share', self.share.id),
            ('availability_zone', self.share.availability_zone),
            ('scheduler_hint', {'only_host': 'host1@backend1#pool1'}),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_replica_create_share_network(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            "2.72"
        )

        arglist = [
            self.share.id,
            '--availability-zone',
            self.share.availability_zone,
            '--share-network',
            self.share.share_network_id,
        ]
        verifylist = [
            ('share', self.share.id),
            ('availability_zone', self.share.availability_zone),
            ('share_network', self.share.share_network_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        if self.share.share_network_id:
            self.replicas_mock.create.assert_called_with(
                share=self.share,
                availability_zone=self.share.availability_zone,
                share_network=self.share.share_network_id,
            )
        else:
            self.replicas_mock.create.assert_called_with(
                share=self.share,
                availability_zone=self.share.availability_zone,
            )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_replica_create_with_metadata(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            "2.95"
        )

        arglist = [
            self.share.id,
            '--availability-zone',
            self.share.availability_zone,
            '--property',
            'fake_key1=fake_value1',
            '--property',
            'fake_key2=fake_value2',
        ]
        verifylist = [
            ('share', self.share.id),
            ('availability_zone', self.share.availability_zone),
            (
                'property',
                {'fake_key1': 'fake_value1', 'fake_key2': 'fake_value2'},
            ),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.replicas_mock.create.assert_called_with(
            share=self.share,
            availability_zone=self.share.availability_zone,
            metadata={'fake_key1': 'fake_value1', 'fake_key2': 'fake_value2'},
        )

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_share_replica_create_metadata_api_version_exception(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            '2.91'
        )
        arglist = [
            self.share.id,
            '--property',
            'fake_key=fake_value',
        ]
        verifylist = [
            ('share', self.share.id),
            ('property', {'fake_key': 'fake_value'}),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_replica_create_wait(self):
        arglist = [self.share.id, '--wait']
        verifylist = [('share', self.share.id), ('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.replicas_mock.create.assert_called_with(
            share=self.share, availability_zone=None
        )

        self.replicas_mock.get.assert_called_with(self.share_replica.id)
        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    @mock.patch('manilaclient.osc.v2.share_replicas.LOG')
    def test_share_replica_create_wait_exception(self, mock_logger):
        arglist = [self.share.id, '--wait']
        verifylist = [('share', self.share.id), ('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_status', return_value=False):
            columns, data = self.cmd.take_action(parsed_args)

            self.replicas_mock.create.assert_called_with(
                share=self.share, availability_zone=None
            )

            mock_logger.error.assert_called_with(
                "ERROR: Share replica is in error state."
            )

            self.replicas_mock.get.assert_called_with(self.share_replica.id)
            self.assertCountEqual(self.columns, columns)
            self.assertCountEqual(self.data, data)


class TestShareReplicaDelete(TestShareReplica):
    def setUp(self):
        super().setUp()

        self.share_replica = manila_fakes.FakeShareReplica.create_one_replica()
        self.replicas_mock.get.return_value = self.share_replica

        self.cmd = osc_share_replicas.DeleteShareReplica(self.app, None)

    def test_share_replica_delete_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_replica_delete(self):
        arglist = [self.share_replica.id]
        verifylist = [('replica', [self.share_replica.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.replicas_mock.delete.assert_called_with(
            self.share_replica, force=False
        )
        self.assertIsNone(result)

    def test_share_replica_delete_force(self):
        arglist = [self.share_replica.id, '--force']
        verifylist = [('replica', [self.share_replica.id]), ('force', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.replicas_mock.delete.assert_called_with(
            self.share_replica, force=True
        )
        self.assertIsNone(result)

    def test_share_replica_delete_multiple(self):
        share_replicas = manila_fakes.FakeShareReplica.create_share_replicas(
            count=2
        )
        arglist = [share_replicas[0].id, share_replicas[1].id]
        verifylist = [
            ('replica', [share_replicas[0].id, share_replicas[1].id])
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertEqual(
            self.replicas_mock.delete.call_count, len(share_replicas)
        )
        self.assertIsNone(result)

    def test_share_snapshot_delete_exception(self):
        arglist = [self.share_replica.id]
        verifylist = [('replica', [self.share_replica.id])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.replicas_mock.delete.side_effect = exceptions.CommandError()
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_replica_delete_wait(self):
        arglist = [self.share_replica.id, '--wait']
        verifylist = [('replica', [self.share_replica.id]), ('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_delete', return_value=True):
            result = self.cmd.take_action(parsed_args)

            self.replicas_mock.delete.assert_called_with(
                self.share_replica, force=False
            )
            self.replicas_mock.get.assert_called_with(self.share_replica.id)
            self.assertIsNone(result)

    def test_share_replica_delete_wait_exception(self):
        arglist = [self.share_replica.id, '--wait']
        verifylist = [('replica', [self.share_replica.id]), ('wait', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch('osc_lib.utils.wait_for_delete', return_value=False):
            self.assertRaises(
                exceptions.CommandError, self.cmd.take_action, parsed_args
            )


class TestShareReplicaList(TestShareReplica):
    columns = [
        'id',
        'status',
        'replica_state',
        'share_id',
        'host',
        'availability_zone',
        'updated_at',
        'properties',
    ]

    column_headers = utils.format_column_headers(columns)

    def setUp(self):
        super().setUp()

        self.share = manila_fakes.FakeShare.create_one_share()
        self.shares_mock.get.return_value = self.share

        self.replicas_list = (
            manila_fakes.FakeShareReplica.create_share_replicas(count=2)
        )
        self.replicas_mock.list.return_value = self.replicas_list

        self.values = (
            oscutils.get_dict_properties(i._info, self.columns)
            for i in self.replicas_list
        )

        self.cmd = osc_share_replicas.ListShareReplica(self.app, None)

    def test_share_replica_list(self):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.replicas_mock.list.assert_called_with(share=None, search_opts={})

        self.assertEqual(self.column_headers, columns)
        self.assertEqual(list(self.values), list(data))

    def test_share_replica_list_for_share(self):
        arglist = ['--share', self.share.id]
        verifylist = [('share', self.share.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.replicas_mock.list.assert_called_with(
            share=self.share, search_opts={}
        )

        self.assertEqual(self.column_headers, columns)
        self.assertEqual(list(self.values), list(data))

    def test_share_replica_list_with_properties_column(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            "2.95"
        )
        metadata = {'replication_policy': 'Sync'}
        self.replicas_list = (
            manila_fakes.FakeShareReplica.create_share_replicas(
                attrs={'metadata': metadata}, count=2
            )
        )
        self.replicas_mock.list.return_value = self.replicas_list

        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.column_headers, columns)

        data_list = list(data)
        self.assertEqual(2, len(data_list))
        for row in data_list:
            props = row[-1]
            self.assertIsInstance(props, format_columns.DictColumn)
            self.assertEqual({'replication_policy': 'Sync'}, props._value)

    def test_share_replica_list_without_properties_column(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            "2.94"
        )
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.assertNotIn('Properties', columns)


class TestShareReplicaShow(TestShareReplica):
    def setUp(self):
        super().setUp()

        self.share_replica = manila_fakes.FakeShareReplica.create_one_replica()
        self.replicas_mock.get.return_value = self.share_replica

        self.replica_el_list = (
            manila_fakes.FakeShareExportLocation.create_share_export_locations(
                count=2
            )
        )

        self.replica_el_mock.list.return_value = self.replica_el_list

        self.cmd = osc_share_replicas.ShowShareReplica(self.app, None)

        self.share_replica._info['export_locations'] = (
            cliutils.convert_dict_list_to_string(self.replica_el_list)
        )

        self.data = tuple(self.share_replica._info.values())
        self.columns = tuple(self.share_replica._info.keys())

    def test_share_replica_show_missing_args(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            osc_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_share_replica_show(self):
        arglist = [self.share_replica.id]
        verifylist = [('replica', self.share_replica.id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.replicas_mock.get.assert_called_with(self.share_replica.id)

        self.assertCountEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestShareReplicaSet(TestShareReplica):
    def setUp(self):
        super().setUp()

        self.share_replica = manila_fakes.FakeShareReplica.create_one_replica(
            methods={"set_metadata": None}
        )
        self.replicas_mock.get.return_value = self.share_replica

        self.cmd = osc_share_replicas.SetShareReplica(self.app, None)

    def test_share_replica_set_replica_state(self):
        new_replica_state = 'in_sync'
        arglist = [self.share_replica.id, '--replica-state', new_replica_state]
        verifylist = [
            ('replica', self.share_replica.id),
            ('replica_state', new_replica_state),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.replicas_mock.reset_replica_state.assert_called_with(
            self.share_replica, new_replica_state
        )
        self.assertIsNone(result)

    def test_share_replica_set_replica_state_exception(self):
        new_replica_state = 'in_sync'
        arglist = [self.share_replica.id, '--replica-state', new_replica_state]
        verifylist = [
            ('replica', self.share_replica.id),
            ('replica_state', new_replica_state),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.replicas_mock.reset_replica_state.side_effect = Exception()

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_replica_set_status(self):
        new_status = 'available'
        arglist = [self.share_replica.id, '--status', new_status]
        verifylist = [
            ('replica', self.share_replica.id),
            ('status', new_status),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.replicas_mock.reset_state.assert_called_with(
            self.share_replica, new_status
        )
        self.assertIsNone(result)

    def test_share_replica_set_status_exception(self):
        new_status = 'available'
        arglist = [self.share_replica.id, '--status', new_status]
        verifylist = [
            ('replica', self.share_replica.id),
            ('status', new_status),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.replicas_mock.reset_state.side_effect = Exception()

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_replica_set_nothing_defined(self):
        arglist = [
            self.share_replica.id,
        ]
        verifylist = [
            ('replica', self.share_replica.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_share_replica_set_property(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            '2.95'
        )
        arglist = [
            self.share_replica.id,
            '--property',
            'fake_key=fake_value',
        ]
        verifylist = [
            ('replica', self.share_replica.id),
            ('property', {'fake_key': 'fake_value'}),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.share_replica.set_metadata.assert_called_once_with(
            {'fake_key': 'fake_value'},
        )

    def test_share_replica_set_property_exception(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            '2.95'
        )
        arglist = [
            self.share_replica.id,
            '--property',
            'fake_key=fake_value',
        ]
        verifylist = [
            ('replica', self.share_replica.id),
            ('property', {'fake_key': 'fake_value'}),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.share_replica.set_metadata.assert_called_once_with(
            {'fake_key': 'fake_value'},
        )

        self.share_replica.set_metadata.side_effect = exceptions.BadRequest
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShareReplicaUnset(TestShareReplica):
    def setUp(self):
        super().setUp()

        self.share_replica = manila_fakes.FakeShareReplica.create_one_replica(
            methods={"delete_metadata": None}
        )
        self.replicas_mock.get.return_value = self.share_replica

        self.cmd = osc_share_replicas.UnsetShareReplica(self.app, None)

    def test_share_replica_unset_property(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            '2.95'
        )
        arglist = [
            '--property',
            'test_key',
            self.share_replica.id,
        ]
        verifylist = [
            ('property', ['test_key']),
            ('replica', self.share_replica.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.share_replica.delete_metadata.assert_called_once_with(
            parsed_args.property
        )

    def test_share_replica_unset_multiple_properties(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            '2.95'
        )
        arglist = [
            '--property',
            'key1',
            '--property',
            'key2',
            self.share_replica.id,
        ]
        verifylist = [
            ('property', ['key1', 'key2']),
            ('replica', self.share_replica.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        # Assert delete_metadata was called for each key
        expected_calls = [mock.call(['key1']), mock.call(['key2'])]
        self.share_replica.delete_metadata.assert_has_calls(
            expected_calls, any_order=False
        )
        self.assertEqual(self.share_replica.delete_metadata.call_count, 2)

    def test_share_replica_unset_property_exception(self):
        self.app.client_manager.share.api_version = api_versions.APIVersion(
            '2.95'
        )
        arglist = [
            self.share_replica.id,
            '--property',
            'test_key',
        ]
        verifylist = [
            ('replica', self.share_replica.id),
            ('property', ['test_key']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # 404 Not Found would be raised, if property 'test_key' doesn't exist.
        self.share_replica.delete_metadata.side_effect = exceptions.NotFound
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShareReplicaPromote(TestShareReplica):
    def setUp(self):
        super().setUp()

        self.share_replica = manila_fakes.FakeShareReplica.create_one_replica(
            attrs={'status': 'available'}
        )
        self.replicas_mock.get.return_value = self.share_replica

        self.cmd = osc_share_replicas.PromoteShareReplica(self.app, None)

    def test_share_replica_promote(self):
        arglist = [
            self.share_replica.id,
        ]
        verifylist = [('replica', self.share_replica.id), ('wait', False)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.replicas_mock.promote.assert_called_with(self.share_replica)
        self.assertIsNone(result)

    def test_share_replica_promote_quiesce_wait_time(self):
        wait_time = '5'
        arglist = [self.share_replica.id, '--quiesce-wait-time', wait_time]
        verifylist = [
            ('replica', self.share_replica.id),
            ('quiesce_wait_time', wait_time),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.replicas_mock.promote.assert_called_with(
            self.share_replica, wait_time
        )
        self.assertIsNone(result)

    @mock.patch.object(
        osc_share_replicas.osc_utils, 'wait_for_status', mock.Mock()
    )
    def test_share_replica_promote_wait(self):
        arglist = [self.share_replica.id, '--wait']
        verifylist = [('replica', self.share_replica.id), ('wait', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.replicas_mock.promote.assert_called_with(self.share_replica)
        self.assertIsNone(result)
        osc_share_replicas.osc_utils.wait_for_status.assert_called_once_with(
            status_f=self.replicas_mock.get,
            res_id=self.share_replica.id,
            success_status=['active'],
            status_field='replica_state',
        )

    def test_share_replica_promote_exception(self):
        arglist = [
            self.share_replica.id,
        ]
        verifylist = [
            ('replica', self.share_replica.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.replicas_mock.promote.side_effect = Exception()

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestShareReplicaResync(TestShareReplica):
    def setUp(self):
        super().setUp()

        self.share_replica = manila_fakes.FakeShareReplica.create_one_replica()
        self.replicas_mock.get.return_value = self.share_replica

        self.cmd = osc_share_replicas.ResyncShareReplica(self.app, None)

    def test_share_replica_resync(self):
        arglist = [
            self.share_replica.id,
        ]
        verifylist = [
            ('replica', self.share_replica.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.replicas_mock.resync.assert_called_with(self.share_replica)
        self.assertIsNone(result)

    def test_share_replica_resync_exception(self):
        arglist = [
            self.share_replica.id,
        ]
        verifylist = [
            ('replica', self.share_replica.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.replicas_mock.resync.side_effect = Exception()

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from manilaclient import config
from manilaclient.tests.functional.osc import base
from tempest.lib.common.utils import data_utils

CONF = config.CONF


class ShareReplicasCLITest(base.OSCClientTestBase):
    def _create_share_and_replica(self, add_cleanup=True, properties=None):
        replication_type = CONF.replication_type
        share_type = self.create_share_type(
            data_utils.rand_name('test_share_type'),
            dhss=True,
            extra_specs={'replication_type': replication_type},
        )
        share_network = self.create_share_network(name='test_share_network')
        share = self.create_share(
            share_type=share_type['name'], share_network=share_network['id']
        )
        replica = self.create_share_replica(
            share['id'],
            share_network=share_network['id'],
            wait=True,
            properties=properties,
            add_cleanup=add_cleanup,
        )
        return replica

    def test_share_replica_create(self):
        share_replica = self._create_share_and_replica()
        share_replica_list = self.listing_result('share replica', 'list')
        self.assertTableStruct(
            share_replica_list,
            [
                'ID',
                'Status',
                'Share ID',
            ],
        )
        self.assertIn(
            share_replica['id'], [item['ID'] for item in share_replica_list]
        )

    def test_share_replica_delete(self):
        share_replica = self._create_share_and_replica(add_cleanup=False)
        self.openstack(f'share replica delete {share_replica["id"]}')
        self.check_object_deleted('share replica', share_replica["id"])

    def test_share_replica_create_with_metadata(self):
        """Create a replica with --property and verify it appears in show."""
        share_replica = self._create_share_and_replica(
            properties={'test_key': 'test_value'}
        )
        show_result = self.dict_result(
            'share', f'replica show {share_replica["id"]}'
        )
        self.assertEqual(share_replica['id'], show_result['id'])
        self.assertIn('test_key', show_result['properties'])
        self.assertIn('test_value', show_result['properties'])

    def test_share_replica_set_property(self):
        """Set a property on a replica and verify it appears in show."""
        share_replica = self._create_share_and_replica()

        self.openstack(
            f'share replica set {share_replica["id"]}'
            ' --property custom_role=secondary'
        )

        show_result = self.dict_result(
            'share', f'replica show {share_replica["id"]}'
        )
        self.assertEqual(share_replica['id'], show_result['id'])
        self.assertIn('custom_role', show_result['properties'])
        self.assertIn('secondary', show_result['properties'])

    def test_share_replica_unset_property(self):
        """Unset a property from a replica and verify it is removed."""
        share_replica = self._create_share_and_replica(
            properties={
                'custom_role': 'secondary',
                'custom_policy': 'async',
            }
        )

        self.openstack(
            f'share replica unset {share_replica["id"]} --property custom_role'
        )

        show_result = self.dict_result(
            'share', f'replica show {share_replica["id"]}'
        )
        self.assertEqual(share_replica['id'], show_result['id'])
        self.assertNotIn('custom_role', show_result['properties'])
        self.assertIn('custom_policy', show_result['properties'])
        self.assertIn('async', show_result['properties'])

    def test_share_replica_list_with_property_filter(self):
        """List replicas filtered by property and verify correct results."""
        share_replica = self._create_share_and_replica(
            properties={'filter_key': 'filter_value'}
        )
        share_id = self.dict_result(
            'share', f'replica show {share_replica["id"]}'
        )['share_id']

        # List filtered by property; the replica should appear
        filtered_list = self.listing_result(
            'share',
            f'replica list --share {share_id}'
            ' --property filter_key=filter_value',
        )
        filtered_ids = [item['ID'] for item in filtered_list]

        self.assertIn(share_replica['id'], filtered_ids)



# Copyright 2015 Mirantis inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manilaclient import api_versions
from manilaclient import base


class ShareExportLocation(base.Resource):
    """Resource class for a share export location."""

    def __repr__(self):
        return f"<ShareExportLocation: {self.id}>"

    def __getitem__(self, key):
        return self._info[key]


class ShareExportLocationManager(base.MetadataCapableManager):
    """Manage :class:`ShareExportLocation` resources."""

    resource_class = ShareExportLocation
    resource_path = '/shares'
    subresource_path = '/export_locations'

    @api_versions.wraps("2.9")
    def list(self, share, search_opts=None):
        """List all share export locations."""
        share_id = base.getid(share)
        return self._list(
            f"/shares/{share_id}/export_locations", "export_locations"
        )

    @api_versions.wraps("2.9")
    def get(self, share, export_location):
        """Get a share export location."""
        share_id = base.getid(share)
        export_location_id = base.getid(export_location)
        return self._get(
            f"/shares/{share_id}/export_locations/{export_location_id}",
            "export_location",
        )

    @api_versions.wraps('2.87')
    def get_metadata(self, share, share_export_location):
        return super().get_metadata(share, subresource=share_export_location)

    @api_versions.wraps('2.87')
    def set_metadata(self, resource, metadata, subresource=None):
        return super().set_metadata(
            resource, metadata, subresource=subresource
        )

    @api_versions.wraps('2.87')
    def delete_metadata(self, resource, keys, subresource=None):
        return super().delete_metadata(resource, keys, subresource=subresource)

    @api_versions.wraps('2.87')
    def update_all_metadata(self, resource, metadata, subresource=None):
        return super().update_all_metadata(
            resource, metadata, subresource=subresource
        )



# Copyright 2012 NetApp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Interface for shares extension."""

from manilaclient import api_versions
from manilaclient import base
from manilaclient.common import constants


class ShareSnapshot(base.MetadataCapableResource):
    """Represent a snapshot of a share."""

    def __repr__(self):
        return f"<ShareSnapshot: {self.id}>"

    def update(self, **kwargs):
        """Update this snapshot."""
        self.manager.update(self, **kwargs)

    def reset_state(self, state):
        """Update the snapshot with the provided state."""
        self.manager.reset_state(self, state)

    def delete(self):
        """Delete this snapshot."""
        self.manager.delete(self)

    def force_delete(self):
        """Delete the specified snapshot ignoring its current state."""
        self.manager.force_delete(self)

    def unmanage_snapshot(self):
        """Unmanage this snapshot."""
        self.manager.unmanage(self)

    def allow(self, access_type, access_to):
        """Allow access to a share snapshot."""
        return self.manager.allow(self, access_type, access_to)

    def deny(self, id):
        """Denies access to a share snapshot."""
        return self.manager.deny(self, id)

    def access_list(self):
        return self.manager.access_list(self)


class ShareSnapshotManager(base.MetadataCapableManager):
    """Manage :class:`ShareSnapshot` resources."""

    resource_class = ShareSnapshot
    resource_path = '/snapshots'

    def _do_create(
        self, share, force=False, name=None, description=None, metadata=None
    ):
        """Create a snapshot of the given share.

        :param share_id: The ID of the share to snapshot.
        :param force: If force is True, create a snapshot even if the
                      share is busy. Default is False.
        :param name: Name of the snapshot
        :param description: Description of the snapshot
        :param metadata: dict - optional metadata to set on share creation
        :rtype: :class:`ShareSnapshot`
        """

        metadata = metadata if metadata is not None else dict()
        body = {
            'snapshot': {
                'share_id': base.getid(share),
                'force': force,
                'name': name,
                'description': description,
                'metadata': metadata,
            }
        }
        return self._create('/snapshots', body, 'snapshot')

    @api_versions.wraps("2.0", "2.72")
    def create(self, share, force=False, name=None, description=None):
        return self._do_create(share, force, name, description)

    @api_versions.wraps("2.73")
    def create(  # noqa
        self,
        share,
        force=False,
        name=None,
        description=None,
        metadata=None,
    ):
        return self._do_create(share, force, name, description, metadata)

    @api_versions.wraps("2.12")
    def manage(
        self,
        share,
        provider_location,
        driver_options=None,
        name=None,
        description=None,
    ):
        """Manage an existing share snapshot.

        :param share: The share object.
        :param provider_location: The provider location of
                                  the snapshot on the backend.
        :param driver_options: dict - custom set of key-values.
        :param name: text - name of new snapshot
        :param description: - description for new snapshot
        """
        driver_options = driver_options if driver_options else {}
        body = {
            'share_id': base.getid(share),
            'provider_location': provider_location,
            'driver_options': driver_options,
            'name': name,
            'description': description,
        }
        return self._create(
            '/snapshots/manage', {'snapshot': body}, 'snapshot'
        )

    @api_versions.wraps("2.12")
    def unmanage(self, snapshot):
        """Unmanage a share snapshot.

        :param snapshot: either snapshot object or text with its ID.
        """
        return self._action("unmanage", snapshot)

    def get(self, snapshot):
        """Get a snapshot.

        :param snapshot: The :class:`ShareSnapshot` instance or string with ID
            of snapshot to delete.
        :rtype: :class:`ShareSnapshot`
        """
        snapshot_id = base.getid(snapshot)
        return self._get(f'/snapshots/{snapshot_id}', 'snapshot')

    def list(
        self, detailed=True, search_opts=None, sort_key=None, sort_dir=None
    ):
        """Get a list of snapshots of shares.

        :param search_opts: Search options to filter out shares.
        :param sort_key: Key to be sorted.
        :param sort_dir: Sort direction, should be 'desc' or 'asc'.
        :rtype: list of :class:`ShareSnapshot`
        """
        search_opts = search_opts or {}

        if sort_key is not None:
            if sort_key in constants.SNAPSHOT_SORT_KEY_VALUES:
                search_opts['sort_key'] = sort_key
            else:
                raise ValueError(
                    'sort_key must be one of the following: {}.'.format(
                        ', '.join(constants.SNAPSHOT_SORT_KEY_VALUES)
                    )
                )

        if sort_dir is not None:
            if sort_dir in constants.SORT_DIR_VALUES:
                search_opts['sort_dir'] = sort_dir
            else:
                raise ValueError(
                    'sort_dir must be one of the following: {}.'.format(
                        ', '.join(constants.SORT_DIR_VALUES)
                    )
                )

        query_string = self._build_query_string(search_opts)

        if detailed:
            path = f"/snapshots/detail{query_string}"
        else:
            path = f"/snapshots{query_string}"

        return self._list(path, 'snapshots')

    def delete(self, snapshot):
        """Delete a snapshot of a share.

        :param snapshot: The :class:`ShareSnapshot` to delete.
        """
        self._delete(f"/snapshots/{base.getid(snapshot)}")

    def _do_force_delete(self, snapshot, action_name="force_delete"):
        """Delete the specified snapshot ignoring its current state."""
        return self._action(action_name, base.getid(snapshot))

    @api_versions.wraps("1.0", "2.6")
    def force_delete(self, snapshot):
        return self._do_force_delete(snapshot, "os-force_delete")

    @api_versions.wraps("2.7")  # noqa
    def force_delete(self, snapshot):  # noqa
        return self._do_force_delete(snapshot, "force_delete")

    def update(self, snapshot, **kwargs):
        """Update a snapshot.

        :param snapshot: The :class:`ShareSnapshot` instance or string with ID
            of snapshot to delete.
        :rtype: :class:`ShareSnapshot`
        """
        if not kwargs:
            return

        body = {
            'snapshot': kwargs,
        }
        snapshot_id = base.getid(snapshot)
        return self._update(f"/snapshots/{snapshot_id}", body)

    def _do_reset_state(self, snapshot, state, action_name="reset_status"):
        """Update the specified share snapshot with the provided state."""
        return self._action(action_name, snapshot, {"status": state})

    @api_versions.wraps("1.0", "2.6")
    def reset_state(self, snapshot, state):
        return self._do_reset_state(snapshot, state, "os-reset_status")

    @api_versions.wraps("2.7")  # noqa
    def reset_state(self, snapshot, state):  # noqa
        return self._do_reset_state(snapshot, state, "reset_status")

    def _do_allow(self, snapshot, access_type, access_to):
        access_params = {
            'access_type': access_type,
            'access_to': access_to,
        }

        return self._action('allow_access', snapshot, access_params)[1][
            'snapshot_access'
        ]

    @api_versions.wraps("2.32")
    def allow(self, snapshot, access_type, access_to):
        return self._do_allow(snapshot, access_type, access_to)

    def _do_deny(self, snapshot, id):
        return self._action('deny_access', snapshot, {'access_id': id})

    @api_versions.wraps("2.32")
    def deny(self, snapshot, id):
        return self._do_deny(snapshot, id)

    def _do_access_list(self, snapshot):
        snapshot_id = base.getid(snapshot)
        access_list = self._list(
            f"/snapshots/{snapshot_id}/access-list", 'snapshot_access_list'
        )
        return access_list

    @api_versions.wraps("2.32")
    def access_list(self, snapshot):
        return self._do_access_list(snapshot)

    def _action(self, action, snapshot, info=None, **kwargs):
        """Perform a snapshot 'action'."""
        body = {action: info}
        self.run_hooks('modify_body_for_action', body, **kwargs)
        url = f'/snapshots/{base.getid(snapshot)}/action'
        return self.api.client.post(url, body=body)



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import os

from oslo_config import cfg
import oslo_log._options as log_options

from manilaclient import api_versions

# 1. Define opts

# "auth_opts" are used by functional tests that are located in
# directory "%project_root%/manilaclient/tests/functional"
auth_opts = [
    # Options for user with 'member' role.
    cfg.StrOpt(
        "username",
        help="This should be the username of a user WITHOUT "
        "administrative privileges.",
    ),
    cfg.StrOpt(
        "tenant_name", help="The non-administrative user's tenant name."
    ),
    cfg.StrOpt(
        "password", secret=True, help="The non-administrative user's password."
    ),
    cfg.StrOpt(
        "auth_url",
        help="URL for where to find the OpenStack Identity public "
        "API endpoint.",
    ),
    cfg.StrOpt(
        "project_domain_name",
        help=(
            "Project domain Name of user with 'member' role "
            "as specified for auth v3."
        ),
    ),
    cfg.StrOpt(
        "project_domain_id",
        help=(
            "Project domain ID of user with 'member' role "
            "as specified for auth v3."
        ),
    ),
    cfg.StrOpt(
        "user_domain_name",
        help=(
            "User domain Name of user with 'member' role "
            "as specified for auth v3."
        ),
    ),
    cfg.StrOpt(
        "user_domain_id",
        help=(
            "User domain ID of user with 'member' role "
            "as specified for auth v3."
        ),
    ),
    # Options for user with 'admin' role.
    cfg.StrOpt(
        "admin_username",
        help="This should be the username of a user WITH "
        "administrative privileges.",
    ),
    cfg.StrOpt(
        "admin_tenant_name", help="The administrative user's tenant name."
    ),
    cfg.StrOpt(
        "admin_password",
        secret=True,
        help="The administrative user's password.",
    ),
    cfg.StrOpt(
        "admin_auth_url",
        help="URL for where to find the OpenStack Identity admin "
        "API endpoint.",
    ),
    cfg.StrOpt(
        "admin_project_domain_name",
        help=(
            "Project domain Name of user with 'admin' role "
            "as specified for auth v3."
        ),
    ),
    cfg.StrOpt(
        "admin_project_domain_id",
        help=(
            "Project domain ID of user with 'admin' role "
            "as specified for auth v3."
        ),
    ),
    cfg.StrOpt(
        "admin_user_domain_name",
        help=(
            "User domain Name of user with 'admin' role "
            "as specified for auth v3."
        ),
    ),
    cfg.StrOpt(
        "admin_user_domain_id",
        help=(
            "User domain ID of user with 'admin' role "
            "as specified for auth v3."
        ),
    ),
    # Other auth options
    cfg.BoolOpt(
        "insecure", default=False, help="Disable SSL certificate verification."
    ),
]

base_opts = [
    cfg.StrOpt(
        "manila_exec_dir",
        default=os.environ.get(
            'OS_MANILA_EXEC_DIR',
            os.path.join(os.path.abspath('.'), '.tox/functional/bin'),
        ),
        help="The path to manilaclient to be executed.",
    ),
    cfg.BoolOpt(
        "suppress_errors_in_cleanup",
        default=True,
        help="Whether to suppress errors with clean up operation or not.",
    ),
]

share_opts = [
    cfg.StrOpt(
        "min_api_microversion",
        default="1.0",
        help="The minimum API microversion is configured to be the "
        "value of the minimum microversion supported by "
        "Manilaclient functional tests. Defaults to 1.0.",
    ),
    cfg.StrOpt(
        "max_api_microversion",
        default=api_versions.MAX_VERSION,
        help="The maximum API microversion is configured to be the "
        "value of the latest microversion supported by "
        "Manilaclient functional tests. Defaults to "
        "manilaclient's max supported API microversion.",
    ),
    cfg.StrOpt(
        "share_network",
        help="Share network Name or ID, that will be used for shares. "
        "Some backend drivers require a share network for share "
        "creation.",
    ),
    cfg.StrOpt(
        "share_network_subnet",
        help="Share network subnet ID. Some backend drivers require a "
        "share network for share creation.",
    ),
    cfg.StrOpt(
        "admin_share_network",
        help="Share network Name or ID, that will be used for shares "
        "in admin tenant.",
    ),
    cfg.StrOpt(
        "share_type",
        help="Share type Name or ID, that will be used with share "
        "creation scheduling. Optional.",
    ),
    cfg.ListOpt(
        "enable_protocols",
        default=["nfs", "cifs"],
        help="List of all enabled protocols. The first protocol in "
        "the list will be used as the default protocol.",
    ),
    cfg.IntOpt(
        "build_interval",
        default=3,
        help="Time in seconds between share availability checks.",
    ),
    cfg.IntOpt(
        "build_timeout",
        default=500,
        help="Timeout in seconds to wait for a share to become available.",
    ),
    cfg.DictOpt(
        'access_types_mapping',
        default={'nfs': 'ip', 'cifs': 'ip'},
        help="Dict contains access types mapping to share "
        "protocol. It will be used to create access rules "
        "for shares. Format: '<protocol>: <type1> <type2>',..."
        "Allowed share protocols: nfs, cifs, cephfs, glusterfs, "
        "hdfs.",
    ),
    cfg.DictOpt(
        'access_levels_mapping',
        default={'nfs': 'rw ro', 'cifs': 'rw'},
        help="Dict contains access levels mapping to share "
        "protocol. It will be used to create access rules for "
        "shares. Format: '<protocol>: <level1> <level2>',... "
        "Allowed share protocols: nfs, cifs, cephfs, glusterfs, "
        "hdfs.",
    ),
    cfg.StrOpt(
        "username_for_user_rules",
        default="stack",
        help="Username, that will be used in share access tests for "
        "user type of access.",
    ),
    cfg.StrOpt(
        "replication_type",
        default="readable",
        choices=["readable", "writable", "dr"],
        help="Replication type to be used when running replication "
        "tests. This option is ignored if run_replication_tests "
        "is set to False.",
    ),
    cfg.BoolOpt(
        "run_replication_tests",
        default=True,
        help="Defines whether to run tests for share replication "
        "or not. Disable this feature if manila driver used "
        "doesn't support share replication.",
    ),
    cfg.BoolOpt(
        "run_snapshot_tests",
        default=True,
        help="Defines whether to run tests that use share snapshots "
        "or not. Disable this feature if used driver doesn't "
        "support it.",
    ),
    cfg.BoolOpt(
        "run_share_servers_tests",
        default=True,
        help="Defines whether to run tests that use share servers "
        "or not. Disable this feature if used driver doesn't "
        "support it or when autodeletion of share servers "
        "is enabled.",
    ),
    cfg.BoolOpt(
        "run_migration_tests",
        default=False,
        help="Defines whether to run share migration tests or not. "
        "Disable this feature if there is no more than one "
        "storage pool being tested or if used driver does not "
        "support it.",
    ),
    cfg.BoolOpt(
        "run_mount_snapshot_tests",
        default=False,
        help="Defines whether to run mountable snapshots tests or "
        "not. Disable this feature if used driver doesn't "
        "support it.",
    ),
    cfg.BoolOpt(
        "run_manage_tests",
        default=False,
        help="Defines whether to run manage/unmanage tests or "
        "not. Disable this feature if used driver does not "
        "support it.",
    ),
    cfg.BoolOpt(
        "run_share_servers_migration_tests",
        default=False,
        help="Defines whether to run share server migration tests or "
        "not. Disable this feature if used driver does not "
        "support it.",
    ),
]

# 2. Generate config

PROJECT_NAME = 'manilaclient'

DEFAULT_CONFIG_FILE = (
    os.environ.get(f'OS_{PROJECT_NAME.upper()}_CONFIG_FILE')
    or f'{PROJECT_NAME}.conf'
)
DEFAULT_CONFIG_DIR = os.environ.get(
    f'OS_{PROJECT_NAME.upper()}_CONFIG_DIR'
) or os.path.join(
    os.path.abspath(os.path.dirname(os.path.dirname(__file__))),
    "etc/manilaclient",
)
DEFAULT_CONFIG_PATH = os.path.join(DEFAULT_CONFIG_DIR, DEFAULT_CONFIG_FILE)
FAILOVER_CONFIG_PATH = f'/etc/{PROJECT_NAME}/{DEFAULT_CONFIG_FILE}'
CONFIG_FILES = []

if os.path.isfile(DEFAULT_CONFIG_PATH):
    CONFIG_FILES.append(DEFAULT_CONFIG_PATH)
if os.path.isfile(FAILOVER_CONFIG_PATH):
    CONFIG_FILES.append(FAILOVER_CONFIG_PATH)

CONF = cfg.CONF

if CONFIG_FILES:
    CONF([], project=PROJECT_NAME, default_config_files=CONFIG_FILES)
else:
    CONF([], project=PROJECT_NAME)

# 3. Register opts

CONF.register_opts(auth_opts)
CONF.register_opts(base_opts)
CONF.register_opts(share_opts)

# 4. Define list_opts for config sample generator


def list_opts():
    """Return a list of oslo_config options available in Manilaclient."""
    opts = [
        (None, copy.deepcopy(auth_opts)),
        (None, copy.deepcopy(base_opts)),
        (None, copy.deepcopy(share_opts)),
    ]
    opts.extend(log_options.list_opts())
    return opts



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging

from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils as oscutils

from manilaclient import api_versions
from manilaclient.common._i18n import _
from manilaclient.common.apiclient import utils as apiutils
from manilaclient.osc import utils

ACCESS_RULE_ATTRIBUTES = [
    'id',
    'share_id',
    'access_level',
    'access_to',
    'access_type',
    'state',
    'access_key',
    'created_at',
    'updated_at',
    'properties',
]

LOG = logging.getLogger(__name__)


class ShareAccessAllow(command.ShowOne):
    """Create a new share access rule."""

    _description = _("Create new share access rule")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share',
            metavar="<share>",
            help=_('Name or ID of the NAS share to modify.'),
        )
        parser.add_argument(
            'access_type',
            metavar="<access_type>",
            help=_(
                'Access rule type (only "ip", "user" (user or group), '
                '"cert" or "cephx" are supported).'
            ),
        )
        parser.add_argument(
            'access_to',
            metavar="<access_to>",
            help=_('Value that defines access.'),
        )
        # metadata --> properties in osc
        parser.add_argument(
            '--properties',
            type=str,
            nargs='*',
            metavar='<key=value>',
            help=_(
                'Space separated list of key=value pairs of properties. '
                'OPTIONAL: Default=None. '
                'Available only for API microversion >= 2.45.'
            ),
        )
        parser.add_argument(
            '--access-level',
            metavar="<access_level>",
            type=str,
            default=None,
            choices=['rw', 'ro'],
            help=_(
                'Share access level ("rw" and "ro" access levels '
                'are supported). Defaults to rw.'
            ),
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            help=_("Wait for share access rule creation."),
        )
        parser.add_argument(
            "--lock-visibility",
            action='store_true',
            default=False,
            help=_(
                "Whether the sensitive fields of the access rule redacted "
                "to other users. Only available with API version >= 2.82."
            ),
        )
        parser.add_argument(
            "--lock-deletion",
            action='store_true',
            default=False,
            help=_(
                "When enabled, a 'delete' lock will be placed against the "
                "rule and the rule cannot be deleted while the lock "
                "exists. Only available with API version >= 2.82."
            ),
        )
        parser.add_argument(
            '--lock-reason',
            metavar="<lock_reason>",
            type=str,
            default=None,
            help=_(
                "Reason for locking the access rule. Should only be "
                "provided alongside a deletion or visibility lock. "
                "Only available with API version >= 2.82."
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share = apiutils.find_resource(share_client.shares, parsed_args.share)
        lock_kwargs = {}
        if parsed_args.lock_visibility:
            lock_kwargs['lock_visibility'] = parsed_args.lock_visibility
        if parsed_args.lock_deletion:
            lock_kwargs['lock_deletion'] = parsed_args.lock_deletion
        if parsed_args.lock_reason:
            lock_kwargs['lock_reason'] = parsed_args.lock_reason

        if lock_kwargs and share_client.api_version < api_versions.APIVersion(
            "2.82"
        ):
            raise exceptions.CommandError(
                'Restricted access rules are only available starting '
                'from API version 2.82.'
            )

        if lock_kwargs.get('lock_reason', None) and not (
            lock_kwargs.get('lock_visibility', None)
            or lock_kwargs.get('lock_deletion', None)
        ):
            raise exceptions.CommandError(
                'Lock reason can only be set while locking the deletion or '
                'visibility.'
            )

        properties = {}
        if parsed_args.properties:
            if share_client.api_version >= api_versions.APIVersion("2.45"):
                properties = utils.extract_properties(parsed_args.properties)
            else:
                raise exceptions.CommandError(
                    "Adding properties to access rules is supported only "
                    "with API microversion 2.45 and beyond"
                )
        try:
            share_access_rule = share.allow(
                access_type=parsed_args.access_type,
                access=parsed_args.access_to,
                access_level=parsed_args.access_level,
                metadata=properties,
                **lock_kwargs,
            )
            if parsed_args.wait:
                if not oscutils.wait_for_status(
                    status_f=share_client.share_access_rules.get,
                    res_id=share_access_rule['id'],
                    status_field='state',
                ):
                    LOG.error(_("ERROR: Share access rule is in error state."))

                share_access_rule = oscutils.find_resource(
                    share_client.share_access_rules, share_access_rule['id']
                )._info
            share_access_rule.update(
                {
                    'properties': utils.format_properties(
                        share_access_rule.pop('metadata', {})
                    )
                }
            )
            return (
                ACCESS_RULE_ATTRIBUTES,
                oscutils.get_dict_properties(
                    share_access_rule, ACCESS_RULE_ATTRIBUTES
                ),
            )
        except Exception as e:
            raise exceptions.CommandError(
                f"Failed to create access to share '{share}': {e}"
            )


class ShareAccessDeny(command.Command):
    """Delete a share access rule."""

    _description = _("Delete a share access rule")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share',
            metavar="<share>",
            help=_('Name or ID of the NAS share to modify.'),
        )
        parser.add_argument(
            'id',
            metavar="<id>",
            help=_('ID of the access rule to be deleted.'),
        )
        parser.add_argument(
            "--wait",
            action='store_true',
            default=False,
            help=_("Wait for share access rule deletion"),
        )
        parser.add_argument(
            "--unrestrict",
            action='store_true',
            default=False,
            help=_(
                "Seek access rule deletion despite restrictions. Only "
                "available with API version >= 2.82."
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share = apiutils.find_resource(share_client.shares, parsed_args.share)
        kwargs = {}
        if parsed_args.unrestrict:
            if share_client.api_version < api_versions.APIVersion("2.82"):
                raise exceptions.CommandError(
                    'Restricted access rules are only available starting from '
                    'API version 2.82.'
                )
            kwargs['unrestrict'] = True

        try:
            share.deny(parsed_args.id, **kwargs)
            if parsed_args.wait:
                if not oscutils.wait_for_delete(
                    manager=share_client.share_access_rules,
                    res_id=parsed_args.id,
                ):
                    raise Exception('timed out while waiting for deletion')
        except Exception as e:
            msg = _(
                "Failed to delete share access rule %(id)s for share "
                "%(share)s: %(e)s"
            )
            raise exceptions.CommandError(
                msg
                % {'id': parsed_args.id, 'share': parsed_args.share, 'e': e}
            )


class ListShareAccess(command.Lister):
    """List share access rules."""

    _description = _("List share access rule")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'share', metavar="<share>", help=_('Name or ID of the share.')
        )

        # metadata --> properties in osc
        parser.add_argument(
            '--properties',
            type=str,
            nargs='*',
            metavar='<key=value>',
            default=None,
            help=_(
                'Filters results by properties (key=value). '
                'OPTIONAL: Default=None. '
                'Available only for API microversion >= 2.45'
            ),
        )
        parser.add_argument(
            "--access-type",
            metavar="<access_type>",
            default=None,
            help=_("Filter access rules by the access type."),
        )
        parser.add_argument(
            "--access-key",
            metavar="<access_key>",
            default=None,
            help=_("Filter access rules by the access key."),
        )
        parser.add_argument(
            "--access-to",
            metavar="<access_to>",
            default=None,
            help=_("Filter access rules by the access to field."),
        )
        parser.add_argument(
            "--access-level",
            metavar="<access_level>",
            default=None,
            help=_("Filter access rules by the access level."),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        share = apiutils.find_resource(share_client.shares, parsed_args.share)
        access_type = parsed_args.access_type
        access_key = parsed_args.access_key
        access_to = parsed_args.access_to
        access_level = parsed_args.access_level

        extended_filter_keys = {
            'access_type': access_type,
            'access_key': access_key,
            'access_to': access_to,
            'access_level': access_level,
        }

        if any(
            extended_filter_keys.values()
        ) and share_client.api_version < api_versions.APIVersion("2.82"):
            raise exceptions.CommandError(
                'Filtering access rules by access_type, access_key, access_to '
                'and access_level is available starting from API version '
                '2.82.'
            )

        search_opts = {}
        if share_client.api_version >= api_versions.APIVersion("2.82"):
            for filter_key, filter_value in extended_filter_keys.items():
                if filter_value:
                    search_opts[filter_key] = filter_value

        if share_client.api_version >= api_versions.APIVersion("2.45"):
            if parsed_args.properties:
                search_opts = {
                    'metadata': utils.extract_properties(
                        parsed_args.properties
                    )
                }
            access_list = share_client.share_access_rules.access_list(
                share, search_opts
            )
        elif parsed_args.properties:
            raise exceptions.CommandError(
                "Filtering access rules by properties is supported only "
                "with API microversion 2.45 and beyond."
            )
        else:
            access_list = share.access_list()

        list_of_keys = [
            'ID',
            'Access Type',
            'Access To',
            'Access Level',
            'State',
        ]

        if share_client.api_version >= api_versions.APIVersion("2.21"):
            list_of_keys.append('Access Key')

        if share_client.api_version >= api_versions.APIVersion("2.33"):
            list_of_keys.append('Created At')
            list_of_keys.append('Updated At')

        values = (
            oscutils.get_item_properties(a, list_of_keys) for a in access_list
        )

        return (list_of_keys, values)


class ShowShareAccess(command.ShowOne):
    """Display a share access rule."""

    _description = _(
        "Display a share access rule. "
        "Available for API microversion 2.45 and higher"
    )

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'access_id',
            metavar="<access_id>",
            help=_('ID of the NAS share access rule.'),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        if share_client.api_version >= api_versions.APIVersion("2.45"):
            access_rule = share_client.share_access_rules.get(
                parsed_args.access_id
            )
            access_rule._info.update(
                {
                    'properties': utils.format_properties(
                        access_rule._info.pop('metadata', {})
                    )
                }
            )
            return (
                ACCESS_RULE_ATTRIBUTES,
                oscutils.get_dict_properties(
                    access_rule._info, ACCESS_RULE_ATTRIBUTES
                ),
            )
        else:
            raise exceptions.CommandError(
                "Displaying share access rule details is only available "
                "with API microversion 2.45 and higher."
            )


class SetShareAccess(command.Command):
    """Set properties to share access rule."""

    _description = _(
        "Set properties to share access rule. "
        "Available for API microversion 2.45 and higher"
    )

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'access_id',
            metavar="<access_id>",
            help=_('ID of the NAS share access rule.'),
        )
        # metadata --> properties in osc
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            default={},
            action=parseractions.KeyValueAction,
            help=_(
                'Set a property to this share access rule. '
                '(Repeat option to set multiple properties) '
                'Available only for API microversion >= 2.45.'
            ),
        )
        parser.add_argument(
            "--access-level",
            metavar="<access_level>",
            default=None,
            choices=['rw', 'ro'],
            help=_(
                'Share access level ("rw" and "ro" access levels '
                'are supported) to set.'
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        if (
            parsed_args.property
            and share_client.api_version < api_versions.APIVersion("2.45")
        ):
            raise exceptions.CommandError(
                "Setting properties to access rule is supported only "
                "with API microversion 2.45 and higher"
            )

        if (
            parsed_args.access_level
            and share_client.api_version < api_versions.APIVersion("2.88")
        ):
            raise exceptions.CommandError(
                "Setting access level to access rule is supported only "
                "with API microversion 2.88 and higher"
            )

        access_rule = share_client.share_access_rules.get(
            parsed_args.access_id
        )
        if parsed_args.property:
            try:
                share_client.share_access_rules.set_metadata(
                    access_rule, parsed_args.property
                )
            except Exception as e:
                raise exceptions.CommandError(
                    "Failed to set properties to share access rule with ID "
                    f"'{access_rule.id}': {e}"
                )

        if parsed_args.access_level:
            try:
                share_client.share_access_rules.set_access_level(
                    access_rule, parsed_args.access_level
                )
            except Exception as e:
                raise exceptions.CommandError(
                    "Failed to set access level to share access rule with ID "
                    f"'{access_rule.id}': {e}"
                )


class UnsetShareAccess(command.Command):
    """Unset properties of share access rule."""

    _description = _(
        "Unset properties of share access rule. "
        "Available for API microversion 2.45 and higher"
    )

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'access_id',
            metavar="<access_id>",
            help=_('ID of the NAS share access rule.'),
        )
        # metadata --> properties in osc
        parser.add_argument(
            '--property',
            metavar='<key>',
            action='append',
            help=_(
                'Remove property from share access rule. '
                '(Repeat option to remove multiple properties) '
                'Available only for API microversion >= 2.45.'
            ),
        )
        return parser

    def take_action(self, parsed_args):
        share_client = self.app.client_manager.share

        if share_client.api_version >= api_versions.APIVersion("2.45"):
            access_rule = share_client.share_access_rules.get(
                parsed_args.access_id
            )
            if parsed_args.property:
                try:
                    share_client.share_access_rules.unset_metadata(
                        access_rule, parsed_args.property
                    )
                except Exception as e:
                    raise exceptions.CommandError(
                        "Failed to unset properties for share access rule "
                        f"with ID '{access_rule.id}': {e}"
                    )
            else:
                raise exceptions.CommandError(
                    "Please specify '--property <key>' to unset a property. "
                )

        else:
            raise exceptions.CommandError(
                "Option to unset properties of access rule is available only "
                "for API microversion 2.45 and higher"
            )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions as lib_exc

from manilaclient.tests.functional.osc import base
from manilaclient.tests.functional import utils

LOCK_DETAIL_ATTRIBUTES = [
    'ID',
    'Resource Id',
    'Resource Type',
    'Resource Action',
    'Lock Context',
    'User Id',
    'Project Id',
    'Created At',
    'Updated At',
    'Lock Reason',
]

LOCK_SUMMARY_ATTRIBUTES = [
    'ID',
    'Resource Id',
    'Resource Type',
    'Resource Action',
]


@utils.skip_if_microversion_not_supported('2.82')
class ResourceLockTests(base.OSCClientTestBase):
    """Lock CLI test cases"""

    def setUp(self):
        super().setUp()
        self.share_type = self.create_share_type(
            name=data_utils.rand_name('lock_tests_type')
        )
        self.share = self.create_share(share_type=self.share_type['id'])

    def test_lock_create_show_use_delete(self):
        """Create a deletion lock on share, view it, try it and remove."""
        lock = self.create_resource_lock(
            self.share['id'],
            lock_reason='tigers rule',
            client=self.user_client,
            add_cleanup=False,
        )

        client_user_id = self.openstack(
            "token issue -c user_id -f value", client=self.user_client
        ).strip()
        client_project_id = self.openstack(
            "token issue -c project_id -f value", client=self.user_client
        ).strip()

        self.assertEqual(self.share['id'], lock['resource_id'])
        self.assertEqual('delete', lock['resource_action'])
        self.assertEqual(client_user_id, lock['user_id'])
        self.assertEqual(client_project_id, lock['project_id'])
        self.assertEqual('user', lock['lock_context'])
        self.assertEqual('tigers rule', lock['lock_reason'])

        lock_show = self.dict_result("share", f"lock show {lock['id']}")
        self.assertEqual(lock['id'], lock_show['ID'])
        self.assertEqual(lock['lock_context'], lock_show['Lock Context'])

        # When a deletion lock exists, share deletion must fail
        self.assertRaises(
            lib_exc.CommandFailed,
            self.openstack,
            f"share delete {self.share['id']}",
        )

        # delete the lock, share will be deleted in cleanup stage
        self.openstack(
            f"share lock delete {lock['id']}", client=self.user_client
        )

        self.assertRaises(
            lib_exc.CommandFailed,
            self.openstack,
            f"share lock show {lock['id']}",
        )

    def test_lock_list_filter_paginate(self):
        lock_1 = self.create_resource_lock(
            self.share['id'],
            lock_reason='tigers rule',
            client=self.user_client,
        )
        lock_2 = self.create_resource_lock(
            self.share['id'],
            lock_reason='tigers still rule',
            client=self.user_client,
        )
        lock_3 = self.create_resource_lock(
            self.share['id'],
            lock_reason='admins rule',
            client=self.admin_client,
        )

        locks = self.listing_result(
            'share', f'lock list --resource {self.share["id"]}'
        )

        self.assertEqual(3, len(locks))
        self.assertEqual(
            sorted(LOCK_SUMMARY_ATTRIBUTES), sorted(locks[0].keys())
        )

        locks = self.listing_result(
            'share',
            f'lock list --lock-context user  --resource {self.share["id"]}',
        )
        self.assertEqual(2, len(locks))
        self.assertNotIn(lock_3['id'], [lock['ID'] for lock in locks])

        locks = self.listing_result(
            'share',
            'lock list --lock-context user'
            f' --resource {self.share["id"]}'
            ' --sort-key created_at '
            ' --sort-dir desc '
            ' --limit 1',
        )
        self.assertEqual(1, len(locks))
        self.assertIn(lock_2['id'], [lock['ID'] for lock in locks])
        self.assertNotIn(lock_1['id'], [lock['ID'] for lock in locks])
        self.assertNotIn(lock_3['id'], [lock['ID'] for lock in locks])

    def test_lock_set_unset_lock_reason(self):
        lock = self.create_resource_lock(
            self.share['id'], client=self.user_client
        )
        self.assertEqual('None', lock['lock_reason'])

        self.openstack(
            f'share lock set --lock-reason \'updated reason\' {lock["id"]}'
        )
        lock_show = self.dict_result("share", f"lock show {lock['id']}")
        self.assertEqual('updated reason', lock_show['Lock Reason'])

        self.openstack(f"share lock unset --lock-reason {lock['id']}")
        lock_show = self.dict_result("share", f"lock show {lock['id']}")
        self.assertEqual('None', lock_show['Lock Reason'])

    def test_lock_restrictions(self):
        """A user can't update or delete a lock created by another user."""
        lock = self.create_resource_lock(
            self.share['id'], client=self.admin_client, add_cleanup=False
        )
        self.assertEqual('admin', lock['lock_context'])

        self.assertRaises(
            lib_exc.CommandFailed,
            self.openstack,
            f"share lock set {lock['id']} --reason 'i cannot do this'",
            client=self.user_client,
        )
        self.assertRaises(
            lib_exc.CommandFailed,
            self.openstack,
            f"share lock unset {lock['id']} --reason",
            client=self.user_client,
        )
        self.assertRaises(
            lib_exc.CommandFailed,
            self.openstack,
            f"share lock delete {lock['id']} ",
            client=self.user_client,
        )

        self.openstack(
            f'share lock set --lock-reason "I can do this" {lock["id"]}',
            client=self.admin_client,
        )

        self.openstack(
            f'share lock delete {lock["id"]}', client=self.admin_client
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import ddt

from manilaclient import api_versions
from manilaclient.tests.unit import utils
from manilaclient.tests.unit.v2 import fakes
from manilaclient.v2 import share_backups

FAKE_BACKUP = 'fake_backup'
FAKE_TARGET_SHARE_ID = 'fake_target_share_id'


@ddt.ddt
class ShareBackupsTest(utils.TestCase):
    class _FakeShareBackup:
        id = 'fake_share_backup_id'

    def setUp(self):
        super().setUp()
        microversion = api_versions.APIVersion("2.91")
        self.manager = share_backups.ShareBackupManager(
            fakes.FakeClient(api_version=microversion)
        )

    def test_delete_str(self):
        with mock.patch.object(self.manager, '_delete', mock.Mock()):
            self.manager.delete(FAKE_BACKUP)
            self.manager._delete.assert_called_once_with(
                share_backups.RESOURCE_PATH % FAKE_BACKUP
            )

    def test_delete_obj(self):
        backup = self._FakeShareBackup
        with mock.patch.object(self.manager, '_delete', mock.Mock()):
            self.manager.delete(backup)
            self.manager._delete.assert_called_once_with(
                share_backups.RESOURCE_PATH % backup.id
            )

    def test_get(self):
        with mock.patch.object(self.manager, '_get', mock.Mock()):
            self.manager.get(FAKE_BACKUP)
            self.manager._get.assert_called_once_with(
                share_backups.RESOURCE_PATH % FAKE_BACKUP,
                share_backups.RESOURCE_NAME,
            )

    def test_restore(self):
        with mock.patch.object(self.manager, '_action', mock.Mock()):
            self.manager.restore(FAKE_BACKUP)
            self.manager._action.assert_called_once_with(
                'restore', FAKE_BACKUP, info=None
            )

    def test_restore_to_target(self):
        with mock.patch.object(self.manager, '_action', mock.Mock()):
            self.manager.restore(
                FAKE_BACKUP, target_share_id=FAKE_TARGET_SHARE_ID
            )
            self.manager._action.assert_called_once_with(
                'restore', FAKE_BACKUP, info=FAKE_TARGET_SHARE_ID
            )

    def test_list(self):
        with mock.patch.object(self.manager, '_list', mock.Mock()):
            self.manager.list()
            self.manager._list.assert_called_once_with(
                share_backups.RESOURCES_PATH + '/detail',
                share_backups.RESOURCES_NAME,
            )

    def test_list_with_share(self):
        with mock.patch.object(self.manager, '_list', mock.Mock()):
            self.manager.list(search_opts={'share_id': 'fake_share_id'})
            share_uri = '?share_id=fake_share_id'
            self.manager._list.assert_called_once_with(
                (share_backups.RESOURCES_PATH + '/detail' + share_uri),
                share_backups.RESOURCES_NAME,
            )

    def test_reset_state(self):
        with mock.patch.object(self.manager, '_action', mock.Mock()):
            self.manager.reset_status(FAKE_BACKUP, 'fake_status')
            self.manager._action.assert_called_once_with(
                'reset_status', FAKE_BACKUP, {'status': 'fake_status'}
            )

    def test_update(self):
        backup = self._FakeShareBackup
        with mock.patch.object(self.manager, '_update', mock.Mock()):
            data = dict(name='backup1')
            self.manager.update(backup, **data)
            self.manager._update.assert_called_once_with(
                share_backups.RESOURCE_PATH % backup.id, {'share_backup': data}
            )
